# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ast
import builtins
import os
import pathlib
import subprocess
import sys
import tempfile
import textwrap
from unittest import mock

import pytest
import testtools
from testtools.matchers import (
    Contains,
    Equals,
    FileExists,
    HasLength,
    LessThan,
    MatchesRegex,
    Not,
)

import snapcraft
from snapcraft import repo
from snapcraft.internal import errors
from snapcraft.plugins.v1 import _ros, catkin
from tests import unit

from . import PluginsV1BaseTestCase


class CatkinPluginBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        # Most Catkin test use core.
        self.project._snap_meta.base = "core"

        class props:
            catkin_packages = ["my_package"]
            source_space = "src"
            source_subdir = None
            include_roscore = False
            catkin_cmake_args = []
            underlay = None
            rosinstall_files = None
            recursive_rosinstall = False
            build_attributes = []
            catkin_ros_master_uri = "http://localhost:11311"
            disable_parallel = False

        self.properties = props()
        self.ros_distro = "kinetic"
        self.ros_version = "1"
        self.ubuntu_distro = "xenial"

        patcher = mock.patch("snapcraft.repo.Ubuntu")
        self.ubuntu_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch(
            "snapcraft.plugins.v1.catkin._find_system_dependencies", return_value={}
        )
        self.dependencies_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("snapcraft.plugins.v1._ros.rosdep.Rosdep")
        self.rosdep_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("snapcraft.plugins.v1.catkin._Catkin")
        self.catkin_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("snapcraft.plugins.v1._ros.wstool.Wstool")
        self.wstool_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("snapcraft.plugins.v1._python.Pip")
        self.pip_mock = patcher.start()
        self.addCleanup(patcher.stop)
        self.pip_mock.return_value.list.return_value = {}

    def assert_rosdep_setup(
        self, rosdistro, ros_version, package_path, rosdep_path, ubuntu_distro
    ):
        self.rosdep_mock.assert_has_calls(
            [
                mock.call(
                    ros_distro=rosdistro,
                    ros_version=ros_version,
                    ros_package_path=package_path,
                    rosdep_path=rosdep_path,
                    ubuntu_distro=ubuntu_distro,
                    base="core",
                    target_arch=self.project._get_stage_packages_target_arch(),
                ),
                mock.call().setup(),
            ]
        )

    def assert_wstool_setup(self, package_path, wstool_path):
        self.wstool_mock.assert_has_calls(
            [
                mock.call(package_path, wstool_path, self.project, "core"),
                mock.call().setup(),
            ]
        )

    def assert_pip_setup(self, python_major_version, part_dir, install_dir, stage_dir):
        self.pip_mock.assert_has_calls(
            [
                mock.call(
                    python_major_version=python_major_version,
                    part_dir=part_dir,
                    install_dir=install_dir,
                    stage_dir=stage_dir,
                ),
                mock.call().setup(),
            ]
        )


class CatkinPluginTestCase(CatkinPluginBaseTest):
    def test_schema(self):
        schema = catkin.CatkinPlugin.schema()

        properties = schema["properties"]
        expected = (
            "catkin-packages",
            "source-space",
            "include-roscore",
            "catkin-cmake-args",
            "underlay",
            "rosinstall-files",
            "recursive-rosinstall",
            "catkin-ros-master-uri",
        )
        self.assertThat(properties, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(properties, Contains(prop))

    def test_schema_catkin_packages(self):
        schema = catkin.CatkinPlugin.schema()

        # Check catkin-packages property
        catkin_packages = schema["properties"]["catkin-packages"]
        expected = ("type", "minitems", "uniqueItems", "items")
        self.assertThat(catkin_packages, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(catkin_packages, Contains(prop))
        self.assertThat(catkin_packages["type"], Equals("array"))
        self.assertThat(catkin_packages["minitems"], Equals(1))
        self.assertTrue(catkin_packages["uniqueItems"])
        self.assertThat(catkin_packages["items"], Contains("type"))
        self.assertThat(catkin_packages["items"]["type"], Equals("string"))

    def test_schema_source_space(self):
        schema = catkin.CatkinPlugin.schema()

        # Check source-space property
        source_space = schema["properties"]["source-space"]
        expected = ("type", "default")
        self.assertThat(source_space, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(source_space, Contains(prop))
        self.assertThat(source_space["type"], Equals("string"))
        self.assertThat(source_space["default"], Equals("src"))

    def test_schema_include_roscore(self):
        schema = catkin.CatkinPlugin.schema()

        # Check include-roscore property
        include_roscore = schema["properties"]["include-roscore"]
        expected = ("type", "default")
        self.assertThat(include_roscore, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(include_roscore, Contains(prop))
        self.assertThat(include_roscore["type"], Equals("boolean"))
        self.assertThat(include_roscore["default"], Equals(True))

    def test_schema_catkin_catkin_cmake_args(self):
        schema = catkin.CatkinPlugin.schema()

        # Check catkin-cmake-args property
        catkin_cmake_args = schema["properties"]["catkin-cmake-args"]
        expected = ("type", "default", "minitems", "items")
        self.assertThat(catkin_cmake_args, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(catkin_cmake_args, Contains(prop))
        self.assertThat(catkin_cmake_args["type"], Equals("array"))
        self.assertThat(catkin_cmake_args["default"], Equals([]))
        self.assertThat(catkin_cmake_args["minitems"], Equals(1))
        self.assertThat(catkin_cmake_args["items"], Contains("type"))
        self.assertThat(catkin_cmake_args["items"]["type"], Equals("string"))

    def test_schema_underlay(self):
        schema = catkin.CatkinPlugin.schema()

        # Check underlay property
        underlay = schema["properties"]["underlay"]
        expected = ("type", "properties", "required")
        self.assertThat(underlay, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(underlay, Contains(prop))
        self.assertThat(underlay["type"], Equals("object"))

        underlay_required = underlay["required"]
        expected = ("build-path", "run-path")
        self.assertThat(underlay_required, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(underlay_required, Contains(prop))

        underlay_properties = underlay["properties"]
        expected = ("build-path", "run-path")
        self.assertThat(underlay_properties, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(underlay_properties, Contains(prop))
        underlay_build_path = underlay_properties["build-path"]
        self.assertThat(underlay_build_path, Contains("type"))
        self.assertThat(underlay_build_path["type"], Equals("string"))
        underlay_run_path = underlay_properties["run-path"]
        self.assertThat(underlay_run_path, Contains("type"))
        self.assertThat(underlay_run_path["type"], Equals("string"))

    def test_schema_rosinstall_files(self):
        schema = catkin.CatkinPlugin.schema()

        # Check rosinstall-files property
        rosinstall_files = schema["properties"]["rosinstall-files"]
        expected = ("type", "default", "minitems", "uniqueItems", "items")
        self.assertThat(rosinstall_files, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(rosinstall_files, Contains(prop))
        self.assertThat(rosinstall_files["type"], Equals("array"))
        self.assertThat(rosinstall_files["default"], Equals([]))
        self.assertThat(rosinstall_files["minitems"], Equals(1))
        self.assertTrue(rosinstall_files["uniqueItems"])
        self.assertThat(rosinstall_files["items"], Contains("type"))
        self.assertThat(rosinstall_files["items"]["type"], Equals("string"))

    def test_schema_recursive_rosinstall(self):
        schema = catkin.CatkinPlugin.schema()

        # Check recursive-rosinstall property
        recursive_rosinstall = schema["properties"]["recursive-rosinstall"]
        expected = ("type", "default")
        self.assertThat(recursive_rosinstall, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(recursive_rosinstall, Contains(prop))
        self.assertThat(recursive_rosinstall["type"], Equals("boolean"))
        self.assertThat(recursive_rosinstall["default"], Equals(False))

    def test_schema_catkin_ros_master_uri(self):
        schema = catkin.CatkinPlugin.schema()

        # Check ros-master-uri property
        catkin_ros_master_uri = schema["properties"]["catkin-ros-master-uri"]
        expected = ("type", "default")
        self.assertThat(catkin_ros_master_uri, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(catkin_ros_master_uri, Contains(prop))
        self.assertThat(catkin_ros_master_uri["type"], Equals("string"))
        self.assertThat(
            catkin_ros_master_uri["default"], Equals("http://localhost:11311")
        )

    def test_get_pull_properties(self):
        expected_pull_properties = [
            "catkin-packages",
            "source-space",
            "include-roscore",
            "underlay",
            "rosinstall-files",
            "recursive-rosinstall",
        ]
        actual_pull_properties = catkin.CatkinPlugin.get_pull_properties()

        self.assertThat(
            actual_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, actual_pull_properties)

    def test_get_build_properties(self):
        expected_build_properties = ["catkin-cmake-args"]
        actual_build_properties = catkin.CatkinPlugin.get_build_properties()

        self.assertThat(
            actual_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, actual_build_properties)

    def test_pull_invalid_dependency(self):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        self.dependencies_mock.return_value = {"apt": {"foo"}}

        self.ubuntu_mock.fetch_stage_packages.side_effect = repo.errors.PackageNotFoundError(
            "foo"
        )

        raised = self.assertRaises(catkin.CatkinAptDependencyFetchError, plugin.pull)

        self.assertThat(
            str(raised),
            Equals(
                "Failed to fetch apt dependencies: The package 'foo' was not found."
            ),
        )

    def test_pull_unable_to_resolve_roscore(self):
        self.properties.include_roscore = True
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        # No system dependencies
        self.dependencies_mock.return_value = {}

        self.rosdep_mock.return_value.resolve_dependency.return_value = None

        self.assertRaises(catkin.CatkinCannotResolveRoscoreError, plugin.pull)

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    def test_pull_invalid_underlay(self, generate_setup_mock):
        self.properties.underlay = {
            "build-path": "test-build-path",
            "run-path": "test-run-path",
        }
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        # No system dependencies
        self.dependencies_mock.return_value = {}

        raised = self.assertRaises(errors.SnapcraftEnvironmentError, plugin.pull)

        self.assertThat(
            str(raised),
            MatchesRegex(".*Requested underlay.*does not point to a valid directory"),
        )

    def test_clean_pull(self):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        self.dependencies_mock.return_value = {"apt": {"foo", "bar", "baz"}}

        plugin.pull()
        os.makedirs(plugin._rosdep_path)

        plugin.clean_pull()
        self.assertFalse(os.path.exists(plugin._rosdep_path))

    def test_valid_catkin_workspace_src(self):
        # sourcedir is expected to be the root of the Catkin workspace. Since
        # it contains a 'src' directory, this is a valid Catkin workspace.
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))
        # An exception will be raised if pull can't handle the valid workspace.
        plugin.pull()

    def test_invalid_catkin_workspace_no_src(self):
        # sourcedir is expected to be the root of the Catkin workspace. Since
        # it does not contain a `src` folder and `source-space` is 'src', this
        # should fail.
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        raised = self.assertRaises(catkin.CatkinPackagePathNotFoundError, plugin.pull)

        self.assertThat(
            str(raised),
            Equals(
                "Failed to find package path: {!r}".format(
                    os.path.join(plugin.sourcedir, "src")
                )
            ),
        )

    def test_valid_catkin_workspace_source_space(self):
        self.properties.source_space = "foo"

        # sourcedir is expected to be the root of the Catkin workspace.
        # Normally this would mean it contained a `src` directory, but it can
        # be remapped via the `source-space` key.
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, self.properties.source_space))
        # An exception will be raised if pull can't handle the source space.
        plugin.pull()

    def test_invalid_catkin_workspace_invalid_source_space(self):
        self.properties.source_space = "foo"

        # sourcedir is expected to be the root of the Catkin workspace. Since
        # it does not contain a `src` folder and source_space wasn't
        # specified, this should fail.
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        raised = self.assertRaises(catkin.CatkinPackagePathNotFoundError, plugin.pull)

        self.assertThat(
            str(raised),
            Equals(
                "Failed to find package path: {!r}".format(
                    os.path.join(plugin.sourcedir, self.properties.source_space)
                )
            ),
        )

    def test_invalid_catkin_workspace_invalid_source_space_build_all(self):
        self.properties.source_space = "foo"
        self.properties.catkin_packages = None

        # sourcedir is expected to be the root of the Catkin workspace. Since
        # it does not contain a `src` folder and source_space wasn't
        # specified, this should fail.
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        raised = self.assertRaises(catkin.CatkinPackagePathNotFoundError, plugin.pull)

        self.assertThat(
            str(raised),
            Equals(
                "Failed to find package path: {!r}".format(
                    os.path.join(plugin.sourcedir, self.properties.source_space)
                )
            ),
        )

    def test_invalid_catkin_workspace_invalid_source_no_packages(self):
        """Test that an invalid source space is fine iff no packages."""
        self.properties.source_space = "foo"
        self.properties.catkin_packages = []

        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)

        # Normally pulling should fail, but since there are no packages to
        # build, even an invalid workspace should be okay.
        plugin.pull()

    def test_invalid_catkin_workspace_source_space_same_as_source(self):
        self.properties.source_space = "."

        # sourcedir is expected to be the root of the Catkin workspace. Since
        # source_space was specified to be the same as the root, this should
        # fail.
        self.assertRaises(
            catkin.CatkinWorkspaceIsRootError,
            catkin.CatkinPlugin,
            "test-part",
            self.properties,
            self.project,
        )

    @mock.patch.object(catkin.CatkinPlugin, "run")
    @mock.patch.object(catkin.CatkinPlugin, "run_output", return_value="foo")
    @mock.patch.object(catkin.CatkinPlugin, "_prepare_build")
    @mock.patch.object(catkin.CatkinPlugin, "_finish_build")
    def test_build_multiple(
        self, finish_build_mock, prepare_build_mock, run_output_mock, run_mock
    ):
        self.properties.catkin_packages.append("package_2")

        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        plugin.build()

        class check_pkg_arguments:
            def __init__(self, test):
                self.test = test

            def __eq__(self, args):
                index = args.index("--pkg")
                packages = args[index + 1 : index + 3]
                self.test.assertIn("my_package", packages)
                self.test.assertIn("package_2", packages)
                return True

        run_mock.assert_called_with(check_pkg_arguments(self))

        self.assertFalse(
            self.dependencies_mock.called,
            "Dependencies should have been discovered in the pull() step",
        )

        finish_build_mock.assert_called_once_with()

    def test_use_in_snap_python_rewrites_shebangs(self):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.rosdir, "bin"))

        # Place a few files with bad shebangs, and some files that shouldn't be
        # changed.
        files = [
            {
                "path": os.path.join(plugin.rosdir, "_setup_util.py"),
                "contents": "#!/foo/bar/baz/python",
                "expected": "#!/usr/bin/env python",
            },
            {
                "path": os.path.join(plugin.rosdir, "bin/catkin"),
                "contents": "#!/foo/baz/python",
                "expected": "#!/usr/bin/env python",
            },
            {
                "path": os.path.join(plugin.rosdir, "foo"),
                "contents": "foo",
                "expected": "foo",
            },
        ]

        for file_info in files:
            with open(file_info["path"], "w") as f:
                f.write(file_info["contents"])

        plugin._use_in_snap_python()

        for file_info in files:
            with open(os.path.join(plugin.rosdir, file_info["path"]), "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))

    @mock.patch.object(catkin.CatkinPlugin, "run")
    @mock.patch.object(catkin.CatkinPlugin, "run_output", return_value="foo")
    def test_use_in_snap_python_skips_binarys(self, run_output_mock, run_mock):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(plugin.rosdir)

        # Place a file to be discovered by _use_in_snap_python().
        open(os.path.join(plugin.rosdir, "foo"), "w").close()

        file_mock = mock.mock_open()
        with mock.patch.object(builtins, "open", file_mock):
            # Reading a binary file may throw a UnicodeDecodeError. Make sure
            # that's handled.
            file_mock.return_value.read.side_effect = UnicodeDecodeError(
                "foo", b"bar", 1, 2, "baz"
            )
            # An exception will be raised if the function can't handle the
            # binary file.
            plugin._use_in_snap_python()

    def test_use_in_snap_python_rewrites_10_ros_sh(self):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.rosdir, "etc", "catkin", "profile.d"))

        ros_profile = os.path.join(
            plugin.rosdir, "etc", "catkin", "profile.d", "10.ros.sh"
        )

        # Place 10.ros.sh with an absolute path to python
        with open(ros_profile, "w") as f:
            f.write("/usr/bin/python foo")

        plugin._use_in_snap_python()

        # Verify that the absolute path in 10.ros.sh was rewritten correctly
        with open(ros_profile, "r") as f:
            self.assertThat(
                f.read(),
                Equals("python foo"),
                "The absolute path to python was not replaced as expected",
            )

    def test_use_in_snap_python_rewrites_1_ros_package_path_sh(self):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.rosdir, "etc", "catkin", "profile.d"))

        ros_profile = os.path.join(
            plugin.rosdir, "etc", "catkin", "profile.d", "1.ros_package_path.sh"
        )

        # Place 1.ros_package_path.sh with an absolute path to python
        with open(ros_profile, "w") as f:
            f.write("/usr/bin/python foo")

        plugin._use_in_snap_python()

        # Verify that the absolute path in 1.ros_package_path.sh was rewritten
        # correctly
        with open(ros_profile, "r") as f:
            self.assertThat(
                f.read(),
                Equals("python foo"),
                "The absolute path to python was not replaced as expected",
            )

    def _verify_run_environment(self, plugin):
        python_path = os.path.join(
            plugin.installdir, "usr", "lib", "python2.7", "dist-packages"
        )
        os.makedirs(python_path)

        # Joining and re-splitting to get hacked script in there as well
        environment = "\n".join(plugin.env(plugin.installdir)).split("\n")

        self.assertThat(
            environment,
            Contains("PYTHONPATH={}${{PYTHONPATH:+:$PYTHONPATH}}".format(python_path)),
        )

        self.assertThat(
            environment,
            Contains("ROS_MASTER_URI={}".format(self.properties.catkin_ros_master_uri)),
        )

        self.assertThat(environment, Contains("ROS_HOME=${SNAP_USER_DATA:-/tmp}/ros"))

        self.assertThat(environment, Contains("LC_ALL=C.UTF-8"))

        # Verify that LD_LIBRARY_PATH was set before setup.sh is sourced
        ld_library_path_index = [
            i for i, line in enumerate(environment) if "LD_LIBRARY_PATH" in line
        ][0]
        source_setup_index = [
            i for i, line in enumerate(environment) if "setup.sh" in line
        ][0]
        self.assertThat(ld_library_path_index, LessThan(source_setup_index))

        return environment

    @mock.patch.object(
        catkin.CatkinPlugin, "_source_setup_sh", return_value="test-source-setup.sh"
    )
    @mock.patch.object(catkin.CatkinPlugin, "run_output", return_value="bar")
    def test_run_environment(self, run_mock, source_setup_sh_mock):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)

        environment = self._verify_run_environment(plugin)

        source_setup_sh_mock.assert_called_with(plugin.installdir, None)
        self.assertThat(environment, Contains("test-source-setup.sh"))

    @mock.patch.object(catkin.CatkinPlugin, "run_output", return_value="bar")
    def test_run_environment_with_underlay(self, run_mock):
        self.properties.underlay = {
            "build-path": "test-build-path",
            "run-path": "test-run-path",
        }
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)

        environment = self._verify_run_environment(plugin)

        setup_path = os.path.join(plugin.rosdir, "snapcraft-setup.sh")
        lines_of_interest = [
            "if [ -f {} ]; then".format(setup_path),
            ". {}".format(setup_path),
            "fi",
        ]
        actual_lines = []

        for line in environment:
            line = line.strip()
            if line in lines_of_interest:
                actual_lines.append(line)

        self.assertThat(
            actual_lines,
            Equals(lines_of_interest),
            "Expected snapcraft-setup.sh to be sourced after checking for its "
            "existence",
        )

    @mock.patch.object(
        catkin.CatkinPlugin, "_source_setup_sh", return_value="test-source-setup.sh"
    )
    @mock.patch.object(catkin.CatkinPlugin, "run_output", return_value="bar")
    def test_run_environment_with_catkin_ros_master_uri(
        self, run_mock, source_setup_sh_mock
    ):

        self.properties.catkin_ros_master_uri = "http://rosmaster:11311"
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)

        self._verify_run_environment(plugin)

    def _evaluate_environment(self, predefinition=""):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)

        python_path = os.path.join(
            plugin.installdir, "usr", "lib", "python2.7", "dist-packages"
        )
        os.makedirs(python_path)

        # Save plugin environment off into a file and read the evaluated
        # version back in, thus obtaining the real environment
        with tempfile.NamedTemporaryFile(mode="w+") as f:
            f.write(predefinition)
            f.write("\n".join(["export " + e for e in plugin.env(plugin.installdir)]))
            f.write('python3 -c "import os; print(dict(os.environ))"')
            f.flush()
            return ast.literal_eval(
                subprocess.check_output(["/bin/sh", f.name])
                .decode(sys.getfilesystemencoding())
                .strip()
            )

    def _pythonpath_segments(self, environment):
        # Verify that the environment contains PYTHONPATH, and return its
        # segments as a list.
        self.assertThat(environment, Contains("PYTHONPATH"))
        return environment["PYTHONPATH"].split(":")

    def _list_contains_empty_items(self, item_list):
        empty_items = [i for i in item_list if not i.strip()]
        return len(empty_items) > 0

    def test_pythonpath_if_not_defined(self):
        environment = self._evaluate_environment()
        segments = self._pythonpath_segments(environment)
        self.assertFalse(
            self._list_contains_empty_items(segments),
            "PYTHONPATH unexpectedly contains empty segments: {}".format(
                environment["PYTHONPATH"]
            ),
        )

    def test_pythonpath_if_null(self):
        environment = self._evaluate_environment(
            textwrap.dedent(
                """
            export PYTHONPATH=
        """
            )
        )

        segments = self._pythonpath_segments(environment)
        self.assertFalse(
            self._list_contains_empty_items(segments),
            "PYTHONPATH unexpectedly contains empty segments: {}".format(
                environment["PYTHONPATH"]
            ),
        )

    def test_pythonpath_if_not_empty(self):
        environment = self._evaluate_environment(
            textwrap.dedent(
                """
            export PYTHONPATH=foo
        """
            )
        )

        segments = self._pythonpath_segments(environment)
        self.assertFalse(
            self._list_contains_empty_items(segments),
            "PYTHONPATH unexpectedly contains empty segments: {}".format(
                environment["PYTHONPATH"]
            ),
        )
        self.assertThat(segments, Contains("foo"))

    @mock.patch.object(
        catkin.CatkinPlugin, "_source_setup_sh", return_value="test-source-setup"
    )
    def test_generate_snapcraft_sh(self, source_setup_sh_mock):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)

        plugin._generate_snapcraft_setup_sh("test-root", None)
        source_setup_sh_mock.assert_called_with("test-root", None)
        with open(os.path.join(plugin.rosdir, "snapcraft-setup.sh"), "r") as f:
            self.assertThat(f.read(), Equals("test-source-setup"))

    def test_source_setup_sh(self):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)

        # Make sure $@ is zeroed, then setup.sh sourced, then $@ is restored
        rosdir = os.path.join("test-root", "opt", "ros", self.ros_distro)
        setup_path = os.path.join(rosdir, "setup.sh")
        lines_of_interest = [
            "set --",
            "if [ -f {} ]; then".format(setup_path),
            "set -- --local",
            "_CATKIN_SETUP_DIR={} . {}".format(rosdir, setup_path),
            "fi",
            'eval "set -- $BACKUP_ARGS"',
        ]
        actual_lines = []

        for line in plugin._source_setup_sh("test-root", None).split("\n"):
            line = line.strip()
            if line in lines_of_interest:
                actual_lines.append(line)

        self.assertThat(
            actual_lines,
            Equals(lines_of_interest),
            "Expected ROS's setup.sh to be sourced after args were zeroed, "
            "followed by the args being restored.",
        )

    def test_generate_snapcraft_sh_with_underlay(self):
        self.properties.underlay = {
            "build-path": "test-build-underlay",
            "run-path": "test-run-underlay",
        }
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)

        # Make sure $@ is zeroed, then setup.sh sourced, then $@ is restored
        underlay_setup_path = os.path.join("test-underlay", "setup.sh")
        rosdir = os.path.join("test-root", "opt", "ros", self.ros_distro)
        setup_path = os.path.join(rosdir, "setup.sh")
        lines_of_interest = [
            "set --",
            "if [ -f {} ]; then".format(underlay_setup_path),
            "set -- --local",
            "_CATKIN_SETUP_DIR={} . {}".format("test-underlay", underlay_setup_path),
            "if [ -f {} ]; then".format(setup_path),
            "set -- --local --extend",
            "_CATKIN_SETUP_DIR={} . {}".format(rosdir, setup_path),
            "fi",
            "fi",
            'eval "set -- $BACKUP_ARGS"',
        ]
        actual_lines = []

        plugin._generate_snapcraft_setup_sh("test-root", "test-underlay")
        with open(os.path.join(plugin.rosdir, "snapcraft-setup.sh"), "r") as f:
            for line in f:
                line = line.strip()
                if line in lines_of_interest:
                    actual_lines.append(line)

        self.assertThat(
            actual_lines,
            Equals(lines_of_interest),
            "Expected ROS's setup.sh to be sourced after args were zeroed, "
            "followed by the args being restored.",
        )

    @mock.patch.object(catkin.CatkinPlugin, "run_output", return_value="bar")
    def test_run_environment_no_python(self, run_mock):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)

        python_path = os.path.join(
            plugin.installdir, "usr", "lib", "python2.7", "dist-packages"
        )

        environment = plugin.env(plugin.installdir)

        self.assertFalse(
            "PYTHONPATH={}".format(python_path) in environment, environment
        )

    @mock.patch.object(catkin.CatkinPlugin, "_use_in_snap_python")
    def test_prepare_build(self, use_python_mock):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.rosdir, "test"))

        # Place a few .cmake files with incorrect paths, and some files that
        # shouldn't be changed.
        files = [
            {
                "path": "fooConfig.cmake",
                "contents": '"/usr/lib/foo"',
                "expected": '"{}/usr/lib/foo"'.format(plugin.installdir),
            },
            {
                "path": "anotherConfig.cmake",
                "contents": '"$ENV{SNAPCRAFT_STAGE}/usr/lib/another"',
                "expected": '"{}/usr/lib/another"'.format(plugin.installdir),
            },
            {
                "path": "bar.cmake",
                "contents": '"/usr/lib/bar"',
                "expected": '"/usr/lib/bar"',
            },
            {
                "path": "test/bazConfig.cmake",
                "contents": '"/test/baz;/usr/lib/baz"',
                "expected": '"{0}/test/baz;{0}/usr/lib/baz"'.format(plugin.installdir),
            },
            {"path": "test/quxConfig.cmake", "contents": "qux", "expected": "qux"},
            {
                "path": "test/installedConfig.cmake",
                "contents": '"{}/foo"'.format(plugin.installdir),
                "expected": '"{}/foo"'.format(plugin.installdir),
            },
        ]

        for file_info in files:
            path = os.path.join(plugin.rosdir, file_info["path"])
            with open(path, "w") as f:
                f.write(file_info["contents"])

        plugin._prepare_build()

        self.assertTrue(use_python_mock.called)

        for file_info in files:
            path = os.path.join(plugin.rosdir, file_info["path"])
            with open(path, "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))


class PullNoUnderlayTestCase(CatkinPluginBaseTest):

    underlay = None
    expected_underlay_path = None

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    def test_pull_debian_dependencies(self, generate_setup_mock):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        self.dependencies_mock.return_value = {"apt": {"foo", "bar", "baz"}}

        plugin.pull()

        self.assert_rosdep_setup(
            self.ros_distro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
        )

        self.wstool_mock.assert_not_called()

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                plugin.installdir, self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        # Verify that dependencies were found as expected. TODO: Would really
        # like to use ANY here instead of verifying explicit arguments, but
        # Python issue #25195 won't let me.
        self.assertThat(self.dependencies_mock.call_count, Equals(1))
        self.assertThat(self.dependencies_mock.call_args[0][0], Equals({"my_package"}))

        # Verify that the dependencies were installed
        self.assertThat(
            self.ubuntu_mock.fetch_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=plugin.stage_packages_path,
                        package_names={"bar", "baz", "foo"},
                        base=plugin.project._get_build_base(),
                        target_arch=plugin.project._get_stage_packages_target_arch(),
                    )
                ]
            ),
        )
        self.assertThat(
            self.ubuntu_mock.unpack_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=plugin.stage_packages_path,
                        install_path=pathlib.Path(plugin.installdir),
                    )
                ]
            ),
        )

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    def test_pull_local_dependencies(self, generate_setup_mock):
        self.properties.catkin_packages.append("package_2")

        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        # No system dependencies (only local)
        self.dependencies_mock.return_value = {}

        plugin.pull()

        self.assert_rosdep_setup(
            self.ros_distro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
        )

        self.wstool_mock.assert_not_called()

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                plugin.installdir, self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        # Verify that dependencies were found as expected. TODO: Would really
        # like to use ANY here instead of verifying explicit arguments, but
        # Python issue #25195 won't let me.
        self.assertThat(self.dependencies_mock.call_count, Equals(1))
        self.assertThat(
            self.dependencies_mock.call_args[0][0], Equals({"my_package", "package_2"})
        )

        # Verify that no .deb packages were installed
        self.assertTrue(
            mock.call().unpack(plugin.installdir) not in self.ubuntu_mock.mock_calls
        )

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    def test_pull_with_roscore(self, generate_setup_mock):
        self.properties.include_roscore = True
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        # No system dependencies
        self.dependencies_mock.return_value = {}

        def resolve(package_name):
            if package_name == "ros_core":
                return {"apt": {"ros-core-dependency"}}

        self.rosdep_mock.return_value.resolve_dependency = resolve

        plugin.pull()

        self.assert_rosdep_setup(
            self.ros_distro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
        )

        self.wstool_mock.assert_not_called()

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                plugin.installdir, self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        # Verify that roscore was installed
        self.assertThat(
            self.ubuntu_mock.fetch_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=plugin.stage_packages_path,
                        package_names={"ros-core-dependency"},
                        base=plugin.project._get_build_base(),
                        target_arch=plugin.project._get_stage_packages_target_arch(),
                    )
                ]
            ),
        )
        self.assertThat(
            self.ubuntu_mock.unpack_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=plugin.stage_packages_path,
                        install_path=pathlib.Path(plugin.installdir),
                    )
                ]
            ),
        )

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    def test_pull_with_rosinstall_files(self, generate_setup_mock):
        self.properties.rosinstall_files = ["rosinstall-file"]
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        # No system dependencies
        self.dependencies_mock.return_value = {}

        plugin.pull()

        self.assert_rosdep_setup(
            self.ros_distro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
        )

        self.assert_wstool_setup(
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "wstool"),
        )

        self.wstool_mock.assert_has_calls(
            [
                mock.call().merge(os.path.join(plugin.sourcedir, "rosinstall-file")),
                mock.call().update(),
            ]
        )

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                plugin.installdir, self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        # Verify that no .deb packages were installed
        self.assertTrue(
            mock.call().unpack(plugin.installdir) not in self.ubuntu_mock.mock_calls
        )

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    def test_pull_pip_dependencies(self, generate_setup_mock):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        self.dependencies_mock.return_value = {"pip": {"foo", "bar", "baz"}}

        plugin.pull()

        self.assert_rosdep_setup(
            self.ros_distro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
        )

        self.wstool_mock.assert_not_called()

        self.assert_pip_setup(
            "2", plugin.partdir, plugin.installdir, plugin.project.stage_dir
        )

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                plugin.installdir, self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        # Verify that dependencies were found as expected. TODO: Would really
        # like to use ANY here instead of verifying explicit arguments, but
        # Python issue #25195 won't let me.
        self.assertThat(self.dependencies_mock.call_count, Equals(1))
        self.assertThat(self.dependencies_mock.call_args[0][0], Equals({"my_package"}))

        # Verify that the pip dependencies were installed
        self.pip_mock.return_value.download.assert_called_once_with(
            {"foo", "bar", "baz"}
        )
        self.pip_mock.return_value.install.assert_called_once_with(
            {"foo", "bar", "baz"}
        )


class PullUnderlayTestCase(CatkinPluginBaseTest):

    underlay = {"build-path": "test-build-path", "run-path": "test-run-path"}
    expected_underlay_path = "test-build-path"

    def setUp(self):
        super().setUp()

        # Make the underlay a valid workspace
        if self.underlay:
            os.makedirs(self.underlay["build-path"])
            open(os.path.join(self.underlay["build-path"], "setup.sh"), "w").close()
            self.properties.underlay = self.underlay

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    def test_pull_debian_dependencies(self, generate_setup_mock):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        self.dependencies_mock.return_value = {"apt": {"foo", "bar", "baz"}}

        plugin.pull()

        self.assert_rosdep_setup(
            self.ros_distro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
        )

        self.wstool_mock.assert_not_called()

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                plugin.installdir, self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        # Verify that dependencies were found as expected. TODO: Would really
        # like to use ANY here instead of verifying explicit arguments, but
        # Python issue #25195 won't let me.
        self.assertThat(self.dependencies_mock.call_count, Equals(1))
        self.assertThat(self.dependencies_mock.call_args[0][0], Equals({"my_package"}))

        # Verify that the dependencies were installed
        self.assertThat(
            self.ubuntu_mock.fetch_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=plugin.stage_packages_path,
                        package_names={"bar", "baz", "foo"},
                        base=plugin.project._get_build_base(),
                        target_arch=plugin.project._get_stage_packages_target_arch(),
                    )
                ]
            ),
        )
        self.assertThat(
            self.ubuntu_mock.unpack_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=plugin.stage_packages_path,
                        install_path=pathlib.Path(plugin.installdir),
                    )
                ]
            ),
        )

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    def test_pull_local_dependencies(self, generate_setup_mock):
        self.properties.catkin_packages.append("package_2")

        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        # No system dependencies (only local)
        self.dependencies_mock.return_value = {}

        plugin.pull()

        self.assert_rosdep_setup(
            self.ros_distro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
        )

        self.wstool_mock.assert_not_called()

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                plugin.installdir, self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        # Verify that dependencies were found as expected. TODO: Would really
        # like to use ANY here instead of verifying explicit arguments, but
        # Python issue #25195 won't let me.
        self.assertThat(self.dependencies_mock.call_count, Equals(1))
        self.assertThat(
            self.dependencies_mock.call_args[0][0], Equals({"my_package", "package_2"})
        )

        # Verify that no .deb packages were installed
        self.assertTrue(
            mock.call().unpack(plugin.installdir) not in self.ubuntu_mock.mock_calls
        )

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    def test_pull_with_roscore(self, generate_setup_mock):
        self.properties.include_roscore = True
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        # No system dependencies
        self.dependencies_mock.return_value = {}

        def resolve(package_name):
            if package_name == "ros_core":
                return {"apt": {"ros-core-dependency"}}

        self.rosdep_mock.return_value.resolve_dependency = resolve

        plugin.pull()

        self.assert_rosdep_setup(
            self.ros_distro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
        )

        self.wstool_mock.assert_not_called()

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                plugin.installdir, self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        # Verify that roscore was installed
        self.assertThat(
            self.ubuntu_mock.fetch_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=plugin.stage_packages_path,
                        package_names={"ros-core-dependency"},
                        base=self.project._get_build_base(),
                        target_arch=self.project._get_stage_packages_target_arch(),
                    )
                ]
            ),
        )
        self.assertThat(
            self.ubuntu_mock.unpack_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=plugin.stage_packages_path,
                        install_path=pathlib.Path(plugin.installdir),
                    )
                ]
            ),
        )

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    def test_pull_with_rosinstall_files(self, generate_setup_mock):
        self.properties.rosinstall_files = ["rosinstall-file"]
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        # No system dependencies
        self.dependencies_mock.return_value = {}

        plugin.pull()

        self.assert_rosdep_setup(
            self.ros_distro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
        )

        self.assert_wstool_setup(
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "wstool"),
        )

        self.wstool_mock.assert_has_calls(
            [
                mock.call().merge(os.path.join(plugin.sourcedir, "rosinstall-file")),
                mock.call().update(),
            ]
        )

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                plugin.installdir, self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        # Verify that no .deb packages were installed
        self.assertTrue(
            mock.call().unpack(plugin.installdir) not in self.ubuntu_mock.mock_calls
        )

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    def test_pull_pip_dependencies(self, generate_setup_mock):
        plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        self.dependencies_mock.return_value = {"pip": {"foo", "bar", "baz"}}

        plugin.pull()

        self.assert_rosdep_setup(
            self.ros_distro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
        )

        self.wstool_mock.assert_not_called()

        self.assert_pip_setup(
            "2", plugin.partdir, plugin.installdir, plugin.project.stage_dir
        )

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                plugin.installdir, self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        # Verify that dependencies were found as expected. TODO: Would really
        # like to use ANY here instead of verifying explicit arguments, but
        # Python issue #25195 won't let me.
        self.assertThat(self.dependencies_mock.call_count, Equals(1))
        self.assertThat(self.dependencies_mock.call_args[0][0], Equals({"my_package"}))

        # Verify that the pip dependencies were installed
        self.pip_mock.return_value.download.assert_called_once_with(
            {"foo", "bar", "baz"}
        )
        self.pip_mock.return_value.install.assert_called_once_with(
            {"foo", "bar", "baz"}
        )


@pytest.fixture
def options():
    class Options:
        catkin_packages = ["my_package"]
        source_space = "src"
        source_subdir = None
        include_roscore = False
        catkin_cmake_args = []
        underlay = None
        rosinstall_files = None
        recursive_rosinstall = False
        build_attributes = []
        catkin_ros_master_uri = "http://localhost:11311"
        disable_parallel = False

    return Options()


class TestBuildArgs:

    scenarios = [
        (
            "release without catkin-cmake-args",
            {
                "build_attributes": [],
                "catkin_cmake_args": [],
                "disable_parallel": False,
            },
        ),
        (
            "release with catkin-cmake-args",
            {
                "build_attributes": [],
                "catkin_cmake_args": ["-DFOO"],
                "disable_parallel": False,
            },
        ),
        (
            "debug without catkin-cmake-args",
            {
                "build_attributes": ["debug"],
                "catkin_cmake_args": [],
                "disable_parallel": False,
            },
        ),
        (
            "debug with catkin-cmake-args",
            {
                "build_attributes": ["debug"],
                "catkin_cmake_args": ["-DFOO"],
                "disable_parallel": False,
            },
        ),
        (
            "release without catkin-cmake-args single threaded",
            {"build_attributes": [], "catkin_cmake_args": [], "disable_parallel": True},
        ),
    ]

    @mock.patch("snapcraft.plugins.v1.catkin.CatkinPlugin.run", autospec=True)
    @mock.patch.object(catkin.CatkinPlugin, "run_output", return_value="foo")
    @mock.patch.object(catkin.CatkinPlugin, "_prepare_build")
    @mock.patch.object(catkin.CatkinPlugin, "_finish_build")
    def test_build(
        self,
        finish_build_mock,
        prepare_build_mock,
        run_output_mock,
        run_mock,
        project_core18,
        options,
        build_attributes,
        catkin_cmake_args,
        disable_parallel,
    ):

        options.build_attributes += build_attributes
        options.catkin_cmake_args += catkin_cmake_args
        options.disable_parallel = disable_parallel

        plugin = catkin.CatkinPlugin("test-part", options, project_core18)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        plugin.build()

        prepare_build_mock.assert_called_once_with()

        if "debug" in build_attributes:
            build_type = "Debug"
        else:
            build_type = "Release"

        expected_command = [
            "catkin_make_isolated",
            "--install",
            "--pkg",
            "my_package",
            "--directory",
            plugin.builddir,
            "--source-space",
            f"{plugin.builddir}/src",
            "--install-space",
            f"{plugin.installdir}/opt/ros/melodic",
            "-j{}".format(1 if disable_parallel else 2),
            "--cmake-args",
            f"-DCMAKE_BUILD_TYPE={build_type}",
        ]
        if catkin_cmake_args:
            expected_command += catkin_cmake_args
        run_mock.assert_called_once_with(plugin, expected_command)

        finish_build_mock.assert_called_once_with()

    @mock.patch.object(catkin.CatkinPlugin, "run")
    @mock.patch.object(catkin.CatkinPlugin, "run_output", return_value="foo")
    @mock.patch.object(catkin.CatkinPlugin, "_prepare_build")
    @mock.patch.object(catkin.CatkinPlugin, "_finish_build")
    def test_build_all_packages(
        self,
        finish_build_mock,
        prepare_build_mock,
        run_output_mock,
        run_mock,
        project_core18,
        options,
        build_attributes,
        catkin_cmake_args,
        disable_parallel,
    ):
        options.catkin_packages = None
        options.build_attributes += build_attributes
        options.catkin_cmake_args += catkin_cmake_args
        options.disable_parallel = disable_parallel

        plugin = catkin.CatkinPlugin("test-part", options, project_core18)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        plugin.build()

        prepare_build_mock.assert_called_once_with()

        if "debug" in build_attributes:
            build_type = "Debug"
        else:
            build_type = "Release"

        expected_command = [
            "catkin_make_isolated",
            "--install",
            "--directory",
            plugin.builddir,
            "--source-space",
            f"{plugin.builddir}/src",
            "--install-space",
            f"{plugin.installdir}/opt/ros/melodic",
            "-j{}".format(1 if disable_parallel else 2),
            "--cmake-args",
            f"-DCMAKE_BUILD_TYPE={build_type}",
        ]
        if catkin_cmake_args:
            expected_command += catkin_cmake_args
        assert run_mock.mock_calls[0][1][0] == expected_command

        finish_build_mock.assert_called_once_with()


class FinishBuildNoUnderlayTestCase(CatkinPluginBaseTest):

    underlay = None
    expected_underlay_path = None

    def setUp(self):
        super().setUp()

        self.properties.underlay = self.underlay
        self.plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    @mock.patch.object(catkin.CatkinPlugin, "_use_in_snap_python")
    def test_finish_build_cmake_paths(self, use_python_mock, generate_setup_mock):
        os.makedirs(os.path.join(self.plugin.rosdir, "test"))

        # Place a few .cmake files with incorrect paths, and some files that
        # shouldn't be changed.
        files = [
            {
                "path": "fooConfig.cmake",
                "contents": '"{}/usr/lib/foo"'.format(self.plugin.installdir),
                "expected": '"$ENV{SNAPCRAFT_STAGE}/usr/lib/foo"',
            },
            {
                "path": "bar.cmake",
                "contents": '"/usr/lib/bar"',
                "expected": '"/usr/lib/bar"',
            },
            {
                "path": "test/bazConfig.cmake",
                "contents": '"{0}/test/baz;{0}/usr/lib/baz"'.format(
                    self.plugin.installdir
                ),
                "expected": '"$ENV{SNAPCRAFT_STAGE}/test/baz;'
                '$ENV{SNAPCRAFT_STAGE}/usr/lib/baz"',
            },
            {"path": "test/quxConfig.cmake", "contents": "qux", "expected": "qux"},
            {
                "path": "test/installedConfig.cmake",
                "contents": '"$ENV{SNAPCRAFT_STAGE}/foo"',
                "expected": '"$ENV{SNAPCRAFT_STAGE}/foo"',
            },
        ]

        for file_info in files:
            path = os.path.join(self.plugin.rosdir, file_info["path"])
            with open(path, "w") as f:
                f.write(file_info["contents"])

        self.plugin._finish_build()

        self.assertTrue(use_python_mock.called)

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                "$SNAP", self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        for file_info in files:
            path = os.path.join(self.plugin.rosdir, file_info["path"])
            with open(path, "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))

        # Verify that no sitecustomize.py was generated
        self.assertThat(
            os.path.join(
                self.plugin.installdir, "usr", "lib", "python2.7", "sitecustomize.py"
            ),
            Not(FileExists()),
        )

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    @mock.patch.object(catkin.CatkinPlugin, "run")
    @mock.patch.object(catkin.CatkinPlugin, "run_output", return_value="foo")
    @mock.patch.object(catkin.CatkinPlugin, "_use_in_snap_python")
    def test_finish_build_cmake_prefix_path(
        self, use_python_mock, run_output_mock, run_mock, generate_setup_mock
    ):
        setup_file = os.path.join(self.plugin.rosdir, "_setup_util.py")
        os.makedirs(os.path.dirname(setup_file))

        with open(setup_file, "w") as f:
            f.write(
                "CMAKE_PREFIX_PATH = '{0}/{1};{0}\n".format(
                    self.plugin.rosdir, self.ros_distro
                )
            )

        self.plugin._finish_build()

        self.assertTrue(use_python_mock.called)

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                "$SNAP", self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        expected = "CMAKE_PREFIX_PATH = []\n"

        with open(setup_file, "r") as f:
            self.assertThat(
                f.read(),
                Equals(expected),
                "The absolute path to python or the CMAKE_PREFIX_PATH "
                "was not replaced as expected",
            )

        # Verify that no sitecustomize.py was generated
        self.assertThat(
            os.path.join(
                self.plugin.installdir, "usr", "lib", "python2.7", "sitecustomize.py"
            ),
            Not(FileExists()),
        )

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    @mock.patch.object(catkin.CatkinPlugin, "run")
    @mock.patch.object(catkin.CatkinPlugin, "run_output", return_value="foo")
    @mock.patch.object(catkin.CatkinPlugin, "_use_in_snap_python")
    def test_finish_build_python_sitecustomize(
        self, use_python_mock, run_output_mock, run_mock, generate_setup_mock
    ):
        self.pip_mock.return_value.list.return_value = {"test-package"}

        # Create site.py, indicating that python2 was a stage-package
        site_py_path = os.path.join(
            self.plugin.installdir, "usr", "lib", "python2.7", "site.py"
        )
        os.makedirs(os.path.dirname(site_py_path), exist_ok=True)
        open(site_py_path, "w").close()

        # Also create python2 site-packages, indicating that pip packages were
        # installed.
        os.makedirs(
            os.path.join(self.plugin.installdir, "lib", "python2.7", "site-packages"),
            exist_ok=True,
        )

        self.plugin._finish_build()

        # Verify that sitecustomize.py was generated
        self.assertThat(
            os.path.join(
                self.plugin.installdir, "usr", "lib", "python2.7", "sitecustomize.py"
            ),
            FileExists(),
        )


class FinishBuildUnderlayTestCase(CatkinPluginBaseTest):

    underlay = {"build-path": "test-build-path", "run-path": "test-run-path"}
    expected_underlay_path = "test-run-path"

    def setUp(self):
        super().setUp()

        self.properties.underlay = self.underlay
        self.plugin = catkin.CatkinPlugin("test-part", self.properties, self.project)

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    @mock.patch.object(catkin.CatkinPlugin, "_use_in_snap_python")
    def test_finish_build_cmake_paths(self, use_python_mock, generate_setup_mock):
        os.makedirs(os.path.join(self.plugin.rosdir, "test"))

        # Place a few .cmake files with incorrect paths, and some files that
        # shouldn't be changed.
        files = [
            {
                "path": "fooConfig.cmake",
                "contents": '"{}/usr/lib/foo"'.format(self.plugin.installdir),
                "expected": '"$ENV{SNAPCRAFT_STAGE}/usr/lib/foo"',
            },
            {
                "path": "bar.cmake",
                "contents": '"/usr/lib/bar"',
                "expected": '"/usr/lib/bar"',
            },
            {
                "path": "test/bazConfig.cmake",
                "contents": '"{0}/test/baz;{0}/usr/lib/baz"'.format(
                    self.plugin.installdir
                ),
                "expected": '"$ENV{SNAPCRAFT_STAGE}/test/baz;'
                '$ENV{SNAPCRAFT_STAGE}/usr/lib/baz"',
            },
            {"path": "test/quxConfig.cmake", "contents": "qux", "expected": "qux"},
            {
                "path": "test/installedConfig.cmake",
                "contents": '"$ENV{SNAPCRAFT_STAGE}/foo"',
                "expected": '"$ENV{SNAPCRAFT_STAGE}/foo"',
            },
        ]

        for file_info in files:
            path = os.path.join(self.plugin.rosdir, file_info["path"])
            with open(path, "w") as f:
                f.write(file_info["contents"])

        self.plugin._finish_build()

        self.assertTrue(use_python_mock.called)

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                "$SNAP", self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        for file_info in files:
            path = os.path.join(self.plugin.rosdir, file_info["path"])
            with open(path, "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))

        # Verify that no sitecustomize.py was generated
        self.assertThat(
            os.path.join(
                self.plugin.installdir, "usr", "lib", "python2.7", "sitecustomize.py"
            ),
            Not(FileExists()),
        )

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    @mock.patch.object(catkin.CatkinPlugin, "run")
    @mock.patch.object(catkin.CatkinPlugin, "run_output", return_value="foo")
    @mock.patch.object(catkin.CatkinPlugin, "_use_in_snap_python")
    def test_finish_build_cmake_prefix_path(
        self, use_python_mock, run_output_mock, run_mock, generate_setup_mock
    ):
        setup_file = os.path.join(self.plugin.rosdir, "_setup_util.py")
        os.makedirs(os.path.dirname(setup_file))

        with open(setup_file, "w") as f:
            f.write(
                "CMAKE_PREFIX_PATH = '{0}/{1};{0}\n".format(
                    self.plugin.rosdir, self.ros_distro
                )
            )

        self.plugin._finish_build()

        self.assertTrue(use_python_mock.called)

        # This shouldn't be called unless there's an underlay
        if self.properties.underlay:
            generate_setup_mock.assert_called_once_with(
                "$SNAP", self.expected_underlay_path
            )
        else:
            generate_setup_mock.assert_not_called()

        expected = "CMAKE_PREFIX_PATH = []\n"

        with open(setup_file, "r") as f:
            self.assertThat(
                f.read(),
                Equals(expected),
                "The absolute path to python or the CMAKE_PREFIX_PATH "
                "was not replaced as expected",
            )

        # Verify that no sitecustomize.py was generated
        self.assertThat(
            os.path.join(
                self.plugin.installdir, "usr", "lib", "python2.7", "sitecustomize.py"
            ),
            Not(FileExists()),
        )

    @mock.patch.object(catkin.CatkinPlugin, "_generate_snapcraft_setup_sh")
    @mock.patch.object(catkin.CatkinPlugin, "run")
    @mock.patch.object(catkin.CatkinPlugin, "run_output", return_value="foo")
    @mock.patch.object(catkin.CatkinPlugin, "_use_in_snap_python")
    def test_finish_build_python_sitecustomize(
        self, use_python_mock, run_output_mock, run_mock, generate_setup_mock
    ):
        self.pip_mock.return_value.list.return_value = {"test-package"}

        # Create site.py, indicating that python2 was a stage-package
        site_py_path = os.path.join(
            self.plugin.installdir, "usr", "lib", "python2.7", "site.py"
        )
        os.makedirs(os.path.dirname(site_py_path), exist_ok=True)
        open(site_py_path, "w").close()

        # Also create python2 site-packages, indicating that pip packages were
        # installed.
        os.makedirs(
            os.path.join(self.plugin.installdir, "lib", "python2.7", "site-packages"),
            exist_ok=True,
        )

        self.plugin._finish_build()

        # Verify that sitecustomize.py was generated
        self.assertThat(
            os.path.join(
                self.plugin.installdir, "usr", "lib", "python2.7", "sitecustomize.py"
            ),
            FileExists(),
        )


class FindSystemDependenciesTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.rosdep_mock = mock.MagicMock()
        self.rosdep_mock.get_dependencies.return_value = {"bar"}

        self.catkin_mock = mock.MagicMock()
        exception = catkin.CatkinPackageNotFoundError("foo")
        self.catkin_mock.find.side_effect = exception

    def test_find_system_dependencies_system_only(self):
        self.rosdep_mock.resolve_dependency.return_value = {"apt": {"baz"}}

        self.assertThat(
            catkin._find_system_dependencies(
                {"foo"}, self.rosdep_mock, self.catkin_mock
            ),
            Equals({"apt": {"baz"}}),
        )

        self.rosdep_mock.get_dependencies.assert_called_once_with("foo")
        self.rosdep_mock.resolve_dependency.assert_called_once_with("bar")
        self.catkin_mock.find.assert_called_once_with("bar")

    def test_find_system_dependencies_system_only_no_packages(self):
        self.rosdep_mock.resolve_dependency.return_value = {"apt": {"baz"}}

        self.assertThat(
            catkin._find_system_dependencies(None, self.rosdep_mock, self.catkin_mock),
            Equals({"apt": {"baz"}}),
        )

        self.rosdep_mock.get_dependencies.assert_called_once_with()
        self.rosdep_mock.resolve_dependency.assert_called_once_with("bar")
        self.catkin_mock.find.assert_called_once_with("bar")

    def test_find_system_dependencies_local_only(self):
        self.assertThat(
            catkin._find_system_dependencies(
                {"foo", "bar"}, self.rosdep_mock, self.catkin_mock
            ),
            HasLength(0),
        )

        self.rosdep_mock.get_dependencies.assert_has_calls(
            [mock.call("foo"), mock.call("bar")], any_order=True
        )
        self.rosdep_mock.resolve_dependency.assert_not_called()
        self.catkin_mock.find.assert_not_called()

    def test_find_system_dependencies_satisfied_in_stage(self):
        self.catkin_mock.find.side_effect = None
        self.catkin_mock.find.return_value = "baz"

        self.assertThat(
            catkin._find_system_dependencies(
                {"foo"}, self.rosdep_mock, self.catkin_mock
            ),
            HasLength(0),
        )

        self.rosdep_mock.get_dependencies.assert_called_once_with("foo")
        self.catkin_mock.find.assert_called_once_with("bar")
        self.rosdep_mock.resolve_dependency.assert_not_called()

    def test_find_system_dependencies_mixed(self):
        self.rosdep_mock.get_dependencies.return_value = {"bar", "baz", "qux"}
        self.rosdep_mock.resolve_dependency.return_value = {"apt": {"quux"}}

        def _fake_find(package_name):
            if package_name == "qux":
                return package_name
            raise catkin.CatkinPackageNotFoundError(package_name)

        self.catkin_mock.find.side_effect = _fake_find
        self.assertThat(
            catkin._find_system_dependencies(
                {"foo", "bar"}, self.rosdep_mock, self.catkin_mock
            ),
            Equals({"apt": {"quux"}}),
        )

        self.rosdep_mock.get_dependencies.assert_has_calls(
            [mock.call("foo"), mock.call("bar")], any_order=True
        )
        self.rosdep_mock.resolve_dependency.assert_called_once_with("baz")
        self.catkin_mock.find.assert_has_calls(
            [mock.call("baz"), mock.call("qux")], any_order=True
        )

    def test_find_system_dependencies_missing_local_dependency(self):
        # Setup a dependency on a non-existing package, and it doesn't resolve
        # to a system dependency.'
        exception = _ros.rosdep.RosdepDependencyNotResolvedError("foo")
        self.rosdep_mock.resolve_dependency.side_effect = exception

        raised = self.assertRaises(
            catkin.CatkinInvalidSystemDependencyError,
            catkin._find_system_dependencies,
            {"foo"},
            self.rosdep_mock,
            self.catkin_mock,
        )

        self.assertThat(
            str(raised),
            Equals(
                "Package 'bar' isn't a valid system dependency. Did "
                "you forget to add it to catkin-packages? If not, "
                "add the Ubuntu package containing it to "
                "stage-packages until you can get it into the rosdep "
                "database."
            ),
        )

    def test_find_system_dependencies_raises_if_unsupported_type(self):
        self.rosdep_mock.resolve_dependency.return_value = {"unsupported-type": {"baz"}}

        raised = self.assertRaises(
            catkin.CatkinUnsupportedDependencyTypeError,
            catkin._find_system_dependencies,
            {"foo"},
            self.rosdep_mock,
            self.catkin_mock,
        )

        self.assertThat(
            str(raised),
            Equals(
                "Package 'bar' resolved to an unsupported type of dependency: "
                "'unsupported-type'."
            ),
        )


class HandleRosinstallFilesTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.wstool_mock = mock.MagicMock()

    def test_single_rosinstall_file(self):
        rosinstall_file = os.path.join("source_path", "rosinstall_file")
        catkin._handle_rosinstall_files(self.wstool_mock, [rosinstall_file])
        self.wstool_mock.merge.assert_called_once_with(
            os.path.join("source_path", "rosinstall_file")
        )

    def test_multiple_rosinstall_files(self):
        rosinstall_files = [
            os.path.join("source_path", "file1"),
            os.path.join("source_path", "file2"),
        ]

        catkin._handle_rosinstall_files(self.wstool_mock, rosinstall_files)

        # The order matters here. It should be the same as how they were passed
        self.wstool_mock.merge.assert_has_calls(
            [
                mock.call(os.path.join("source_path", "file1")),
                mock.call(os.path.join("source_path", "file2")),
            ]
        )


class RecursivelyHandleRosinstallFilesTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.wstool_mock = mock.MagicMock()

    def test_recursive_rosinstall(self):
        counter = 0

        # A fake update that plops a new rosinstall file down every time
        # it's called (up to two times).
        def _fake_wstool_update():
            nonlocal counter
            if counter < 2:
                counter += 1
                open(
                    os.path.join("source_path", "{}.rosinstall".format(counter)), "w"
                ).close()

        self.wstool_mock.update.side_effect = _fake_wstool_update

        os.mkdir("source_path")
        open(os.path.join("source_path", "0.rosinstall"), "w").close()

        catkin._recursively_handle_rosinstall_files(self.wstool_mock, "source_path")

        self.wstool_mock.merge.assert_has_calls(
            [
                mock.call(os.path.join("source_path", "0.rosinstall")),
                mock.call(os.path.join("source_path", "1.rosinstall")),
                mock.call(os.path.join("source_path", "2.rosinstall")),
            ]
        )


class CatkinFindTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.project = snapcraft.project.Project()
        self.project._snap_meta.build_base = "core"
        self.catkin = catkin._Catkin(
            "kinetic", "workspace_path", "catkin_path", self.project
        )

        patcher = mock.patch("snapcraft.repo.Ubuntu")
        self.ubuntu_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("subprocess.check_output")
        self.check_output_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def test_setup(self):
        # Return something other than a Mock to ease later assertions
        self.check_output_mock.return_value = b""

        self.catkin.setup()

        # Verify that only rospack was installed (no other .debs)
        self.assertThat(
            self.ubuntu_mock.fetch_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=self.catkin._catkin_stage_packages_path,
                        package_names=["ros-kinetic-catkin"],
                        base=self.project._get_build_base(),
                        target_arch=self.project._get_stage_packages_target_arch(),
                    )
                ]
            ),
        )
        self.assertThat(
            self.ubuntu_mock.unpack_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=self.catkin._catkin_stage_packages_path,
                        install_path=pathlib.Path("catkin_path/install"),
                    )
                ]
            ),
        )

    def test_setup_can_run_multiple_times(self):
        self.catkin.setup()

        # Make sure running setup() again doesn't have problems with the old
        # environment. An exception will be raised if setup() can't be called
        # twice.
        self.catkin.setup()

    def test_find(self):
        self.check_output_mock.return_value = b"bar"

        self.assertThat(self.catkin.find("foo"), Equals("bar"))

        self.assertTrue(self.check_output_mock.called)
        positional_args = self.check_output_mock.call_args[0][0]
        self.assertThat(
            " ".join(positional_args), Contains("catkin_find --first-only foo")
        )

    def test_find_only_in_catkin_workspace(self):
        self.check_output_mock.return_value = os.path.join(
            self.catkin._catkin_install_path, "bar"
        ).encode(sys.getfilesystemencoding())

        with testtools.ExpectedException(
            catkin.CatkinPackageNotFoundError, "Unable to find Catkin package 'foo'"
        ):
            self.catkin.find("foo")

    def test_find_non_existing_package(self):
        self.check_output_mock.side_effect = subprocess.CalledProcessError(1, "foo")

        with testtools.ExpectedException(
            catkin.CatkinPackageNotFoundError, "Unable to find Catkin package 'foo'"
        ):
            self.catkin.find("foo")

        self.assertTrue(self.check_output_mock.called)
        positional_args = self.check_output_mock.call_args[0][0]
        self.assertThat(
            " ".join(positional_args), Contains("catkin_find --first-only foo")
        )


class TestCatkinCmakeArg:
    scenarios = [
        (
            "no quote",
            dict(input_string="-DCMAKE_FOO=BAR", expected_string="-DCMAKE_FOO=BAR"),
        ),
        (
            "no quote multiple",
            dict(
                input_string="-DCMAKE_C_FLAGS=-Wall -Werror",
                expected_string="-DCMAKE_C_FLAGS=-Wall -Werror",
            ),
        ),
        (
            "single quote",
            dict(input_string="-DCMAKE_FOO='BAR'", expected_string="-DCMAKE_FOO=BAR"),
        ),
        (
            "single quote multiple",
            dict(
                input_string="-DCMAKE_C_FLAGS='-Wall -Werror'",
                expected_string="-DCMAKE_C_FLAGS=-Wall -Werror",
            ),
        ),
        (
            "double quote",
            dict(input_string='-DCMAKE_FOO="BAR"', expected_string="-DCMAKE_FOO=BAR"),
        ),
        (
            "double quote multiple",
            dict(
                input_string='-DCMAKE_C_FLAGS="-Wall -Werror"',
                expected_string="-DCMAKE_C_FLAGS=-Wall -Werror",
            ),
        ),
        (
            "nested double quote",
            dict(
                input_string='-DCMAKE_C_FLAGS="-Wall -Werror -DTEXT=\\"Hello Friend\\""',
                expected_string='-DCMAKE_C_FLAGS=-Wall -Werror -DTEXT="Hello Friend"',
            ),
        ),
        (
            "nested single quote",
            dict(
                input_string="-DCMAKE_C_FLAGS=\"-Wall -Werror -DTEXT='Hello Friend'\"",
                expected_string="-DCMAKE_C_FLAGS=-Wall -Werror -DTEXT='Hello Friend'",
            ),
        ),
    ]

    def test_arg(self, input_string, expected_string):
        assert catkin._parse_cmake_arg(input_string) == expected_string



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import pathlib
import re
import shlex
import shutil
from typing import Optional

from snapcraft.internal import common

from . import errors
from ._utils import _executable_is_valid

logger = logging.getLogger(__name__)
_COMMAND_PATTERN = re.compile("^[A-Za-z0-9. _#:$-][A-Za-z0-9/. _#:$-]*$")
_FMT_SNAPD_WRAPPER = (
    "A shell wrapper will be generated for command {!r} as it does not conform "
    "with the command pattern expected by the runtime. "
    "Commands must be relative to the prime directory and can only consist "
    "of alphanumeric characters, spaces, and the following special characters: "
    "/ . _ # : $ -"
)


def _get_shebang_from_file(file_path: pathlib.Path) -> Optional[str]:
    """Get the shebang from specified file."""
    with open(file_path, "rb") as exefile:
        if exefile.read(2) != b"#!":
            return None
        shebang_line = exefile.readline().strip().decode("utf-8")

    return shebang_line


class _SnapCommandResolver:
    def __init__(self, *, command: str, prime_path: pathlib.Path) -> None:
        self.original_command = command
        self._prime_path = prime_path
        self.interpreter: Optional[str] = None
        self.command = command

    def _find_executable(self, command: str) -> Optional[pathlib.Path]:
        """Find executable in common binary path locations.

        :return: Path to executable found, whether it is in searched
            prime directory paths, or on host using shutil.which().
        """
        binary_paths = (
            pathlib.Path(p, command)
            for p in common.get_bin_paths(root=self._prime_path)
        )

        # Final all potential hits, sorting to ensure consistent behavior.
        primed_binary_paths = sorted(
            [p for p in binary_paths if _executable_is_valid(p)]
        )

        # Warn if we have multiple potential hits.
        if len(primed_binary_paths) > 0:
            if len(primed_binary_paths) > 1:
                matches = [
                    str(p.relative_to(self._prime_path)) for p in primed_binary_paths
                ]
                logger.warning(
                    f"Multiple binaries matching for ambiguous command {command!r}: {matches!r}"
                )
            return primed_binary_paths[0]

        # Last chance to find in the prime_dir, mostly for backwards compatibility,
        # to find the executable, historical snaps like those built with the catkin
        # plugin will have roslaunch in a path like /opt/ros/bin/roslaunch.
        for root, _, files in os.walk(self._prime_path):
            file_path = pathlib.Path(root, command)
            if _executable_is_valid(str(file_path)):
                return file_path

        # Finally, check if it is part of the system.
        which_command = shutil.which(command)
        if which_command is not None:
            return pathlib.Path(which_command)

        return None

    def resolve_interpreter(self) -> Optional[str]:
        """Resolve the interpreter to a format suitable for use in a snap.yaml.

        Effectively performs the job of "/usr/bin/env", ignoring
        it if used as the (initial) interpreter.

        :return: Resolved and normalized interpreter path, if any.
            Incorporate arguments found in shebang.
        """
        interpreter = self.get_command_interpreter()
        if interpreter is None:
            return None

        # Remove the leading /usr/bin/env and resolve it now.
        interpreter_parts = shlex.split(interpreter, posix=False)
        if interpreter_parts[0] == "/usr/bin/env":
            interpreter_parts = interpreter_parts[1:]

        # If interpreter is absolute, ignore it.
        if pathlib.Path(interpreter_parts[0]).is_absolute():
            return None

        # Strip leading unnecessary $SNAP from command, if present.
        interpreter_parts[0] = re.sub(r"^\$SNAP/", "", interpreter_parts[0])

        # Check if relative to prime, where it should be found.
        if pathlib.Path(self._prime_path, interpreter_parts[0]).exists():
            return " ".join(interpreter_parts)

        # If not where it claims to be, search for backwards compatibility.
        interpreter_path = self._find_executable(command=interpreter_parts[0])

        if interpreter_path is None:
            # Interpreter was not found, note that this is not a hard error.
            logger.warning(
                f"The interpreter {interpreter_parts[0]!r} for {self.original_command!r} was not found."
            )
        else:
            # Interpreter was found in the prime directory, but required the
            # legacy search, or it was found on the host.  Log a warning.
            if self._prime_path in interpreter_path.parents:
                resolved_interpreter = interpreter_path.relative_to(
                    self._prime_path
                ).as_posix()
            else:
                resolved_interpreter = interpreter_path.as_posix()

            logger.warning(
                f"The interpreter {interpreter_parts[0]!r} for {self.original_command!r} was resolved to {resolved_interpreter!r}."
            )
            interpreter_parts[0] = resolved_interpreter

        return " ".join(interpreter_parts)

    def resolve_command(self) -> Optional[str]:
        """Resolve the given command, searching prime directory or host if needed.

        :return: Resolved and normalized command string.
        """
        command_parts = shlex.split(self.command, posix=False)

        # Strip leading unnecessary $SNAP from command, if present.
        command_parts[0] = re.sub(r"^\$SNAP/", "", command_parts[0])

        # Check if relative to prime, where it should be found.
        if os.path.exists(os.path.join(self._prime_path, command_parts[0])):
            return " ".join(command_parts)

        # If not where it claims to be, search for backwards compatibility.
        command_path = self._find_executable(command=command_parts[0])
        if command_path is None:
            return None

        # Command was found in the prime directory, but it required the
        # legacy search or was found on the host.  Log a warning.
        if self._prime_path in command_path.parents:
            resolved_command = command_path.relative_to(self._prime_path).as_posix()
        else:
            resolved_command = command_path.as_posix()

        logger.warning(
            f"The command {command_parts[0]!r} for {self.original_command!r} was resolved to {resolved_command!r}."
        )
        command_parts[0] = resolved_command

        return " ".join(command_parts)

    def get_command_interpreter(self) -> Optional[str]:
        """Get interpreter (shebang) for executable pointed to by self.command.

        :return: Interpreter string, else None if none found.
        """
        command_parts = shlex.split(self.command, posix=False)

        command_path = pathlib.Path(self._prime_path, command_parts[0])
        if not command_path.exists():
            return None

        return _get_shebang_from_file(command_path)

    def resolve(self) -> Optional[str]:
        """Resolve command to take into account interpreter and pathing,
        resolving ambiguous commands either relative to prime, or absolute
        to host/base.

        :return: Resolved and normalized command, accounting for the interpreter,
          if required.  None if command not found.
        """
        # Strip leading unneed $SNAP from command, if present.
        self.command = re.sub(r"^\$SNAP/", "", self.command)

        # Resolve interpreter, if any.
        self.interpreter = self.resolve_interpreter()

        # Resolve command.
        resolved_command = self.resolve_command()
        if resolved_command is None:
            return None

        self.command = resolved_command

        # Format interpreter and command for snap.yaml.
        if self.interpreter is not None:
            finalized_command = " ".join(
                [self.interpreter, pathlib.Path("$SNAP", self.command).as_posix()]
            )
        else:
            finalized_command = self.command

        # Issue final warnings for any changes.
        if finalized_command != self.original_command:
            if self.original_command.startswith("$SNAP/"):
                logger.warning(
                    f"Found unneeded '$SNAP/' in command {self.original_command!r}."
                )

            if self.interpreter is not None:
                logger.warning(
                    f"The command {self.original_command!r} has been changed to {finalized_command!r} to safely account for the interpreter."
                )
            else:
                logger.warning(
                    f"The command {self.original_command!r} has been changed to {finalized_command!r}."
                )

        return finalized_command

    @classmethod
    def resolve_snap_command_entry(
        cls, *, command: str, prime_path: pathlib.Path
    ) -> Optional[str]:
        """Resolve and normalize a given snap command entry, taking
        the interpreter into account, if appropriate.

        :param command: command to resolve and normalize.
        :param prime_path: path to prime directory.
        :return: Resolved and normalized command string suitable for a snap.yaml
            if no wrapper is required.
        """
        # If command is absolute, nothing to resolve.
        if command.startswith("/"):
            return command
        else:
            resolver = cls(command=command, prime_path=prime_path)
            return resolver.resolve()


class Command:
    """Representation of a command string."""

    def __str__(self) -> str:
        return self.command

    def __init__(self, *, app_name: str, command_name: str, command: str) -> None:
        self._app_name = app_name
        self._command_name = command_name
        self.command = command
        self.wrapped_command: Optional[str] = None
        self.massaged_command: Optional[str] = None

    @property
    def command_name(self) -> str:
        """Read-only to ensure consistency with app dictionary mappings."""
        return self._command_name

    @property
    def requires_wrapper(self) -> bool:
        if self.wrapped_command is not None:
            command = self.wrapped_command
        else:
            command = self.command

        return command.startswith("/") or not _COMMAND_PATTERN.match(command)

    @property
    def wrapped_command_name(self) -> str:
        """Return the relative in-snap path to the wrapper for command."""
        return "{command_name}-{app_name}.wrapper".format(
            command_name=self.command_name, app_name=self._app_name
        )

    def prime_command(
        self, *, can_use_wrapper: bool, massage_command: bool = True, prime_dir: str,
    ) -> str:
        """Finalize and prime command, massaging as necessary.

        Check if command is in prime_dir and raise exception if not valid."""

        if massage_command:
            resolved_command = _SnapCommandResolver.resolve_snap_command_entry(
                command=self.command, prime_path=pathlib.Path(prime_dir)
            )
            if resolved_command is None:
                raise errors.InvalidAppCommandNotFound(
                    command=self.command, app_name=self._app_name
                )
            self.command = resolved_command

        if self.requires_wrapper:
            if not can_use_wrapper:
                raise errors.InvalidAppCommandFormatError(self.command, self._app_name)
            self.wrapped_command = self.command
            if not _COMMAND_PATTERN.match(self.command):
                logger.warning(_FMT_SNAPD_WRAPPER.format(self.command))
            self.command = self.wrapped_command_name
        else:
            command_parts = shlex.split(self.command)
            command_path = os.path.join(prime_dir, command_parts[0])

            if not os.path.exists(command_path):
                raise errors.InvalidAppCommandNotFound(
                    command=self.command, app_name=self._app_name
                )

            if not _executable_is_valid(command_path):
                raise errors.InvalidAppCommandNotExecutable(
                    command=self.command, app_name=self._app_name
                )

        return self.command

    def write_wrapper(self, *, prime_dir: str) -> Optional[str]:
        """Write command wrapper if required for this command."""
        if self.wrapped_command is None:
            return None

        command_wrapper = os.path.join(prime_dir, self.wrapped_command_name)

        # We cannot exec relative paths in our wrappers.
        if self.wrapped_command.startswith("/"):
            command = self.wrapped_command
        else:
            command = os.path.join("$SNAP", self.wrapped_command)

        with open(command_wrapper, "w+") as command_file:
            print("#!/bin/sh", file=command_file)
            print('exec {} "$@"'.format(command), file=command_file)

        os.chmod(command_wrapper, 0o755)
        return command_wrapper



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import enum
import os
from copy import deepcopy
from typing import Any, Dict, List, Optional, Sequence  # noqa: F401

from snapcraft import yaml_utils

from . import errors
from ._utils import _executable_is_valid
from .command import Command
from .desktop import DesktopFile

_COMMAND_ENTRIES = ["command", "stop-command"]
_MASSAGED_BASES = ["core", "core18"]


@enum.unique
class ApplicationAdapter(enum.Enum):
    NONE = 1
    LEGACY = 2
    FULL = 3


class Application:
    """Representation of an app entry in snapcraft.yaml"""

    def __init__(
        self,
        *,
        app_name: str,
        app_properties: Dict[str, Any] = None,
        adapter: ApplicationAdapter,
        desktop: str = None,
        install_mode: str = None,
        command_chain: List[str] = None,
        passthrough: Dict[str, Any] = None,
        commands: Dict[str, Command] = None
    ) -> None:
        """Initialize an application entry.

        TODO: Eliminate use of app_properties.
        """
        self._app_name = app_name

        self._app_properties: Dict[str, Any] = dict()
        if app_properties:
            self._app_properties = app_properties

        self.adapter = adapter
        self.desktop = desktop
        self.install_mode = install_mode

        self.command_chain: List[str] = list()
        if command_chain:
            self.command_chain = command_chain

        self.commands: Dict[str, Command] = dict()
        if commands:
            self.commands = commands

        self.passthrough: Dict[str, Any] = dict()
        if passthrough:
            self.passthrough = passthrough

    @property
    def app_name(self) -> str:
        """Read-only to ensure consistency with Snap dictionary mappings."""
        return self._app_name

    def can_use_wrapper(self, base: Optional[str]) -> bool:
        """Return if an wrapper should be allowed for app entries."""
        # Force use of no wrappers when command-chain is set.
        if self.command_chain:
            return False

        # We only allow wrappers for core and core18.
        if not self._massage_commands(base=base):
            return False

        # Now that command-chain and bases have been checked for,
        # check if the none adapter has been forced.
        if self.adapter == ApplicationAdapter.NONE:
            return False

        return True

    def _massage_commands(self, *, base: Optional[str]) -> bool:
        return base in _MASSAGED_BASES

    def prime_commands(self, *, base: Optional[str], prime_dir: str) -> None:
        can_use_wrapper = self.can_use_wrapper(base)
        massage_command = self._massage_commands(base=base)
        for command in self.commands.values():
            command.prime_command(
                can_use_wrapper=can_use_wrapper,
                massage_command=massage_command,
                prime_dir=prime_dir,
            )

    def write_command_wrappers(self, *, prime_dir: str) -> None:
        for command in self.commands.values():
            command.write_wrapper(prime_dir=prime_dir)

    def write_application_desktop_file(
        self, snap_name: str, prime_dir: str, gui_dir: str, icon_path: Optional[str]
    ) -> None:
        if not self.desktop:
            return

        desktop_file = DesktopFile(
            snap_name=snap_name,
            app_name=self.app_name,
            filename=self.desktop,
            prime_dir=prime_dir,
        )

        desktop_file.write(gui_dir=gui_dir, icon_path=icon_path)

    def validate_command_chain_executables(self, prime_dir: str) -> None:
        for item in self.command_chain:
            executable_path = os.path.join(prime_dir, item)

            # command-chain entries must always be relative to the root of
            # the snap, i.e. PATH is not used.
            if not _executable_is_valid(executable_path):
                raise errors.InvalidCommandChainError(item, self.app_name)

    def validate(self) -> None:
        """Validate application, raising exception on error."""

        # No checks here (yet).
        return

    @classmethod
    def from_dict(cls, *, app_dict: Dict[str, Any], app_name: str) -> "Application":
        """Create application from dictionary."""

        app_dict = deepcopy(app_dict)

        adapter_string = app_dict.get("adapter", "full").upper()
        adapter = ApplicationAdapter[adapter_string]

        app = Application(
            app_name=app_name,
            app_properties=app_dict,
            adapter=adapter,
            desktop=app_dict.get("desktop", None),
            install_mode=app_dict.get("install-mode", None),
            command_chain=app_dict.get("command-chain", None),
            passthrough=app_dict.get("passthrough", None),
        )

        # Populate commands from app_properties.
        for command_name in _COMMAND_ENTRIES:
            if command_name not in app_dict:
                continue

            app.commands[command_name] = Command(
                app_name=app_name,
                command_name=command_name,
                command=app_dict[command_name],
            )

        return app

    def to_dict(self) -> Dict[str, Any]:
        """Returns and ordered dictionary with the transformed app entry."""

        app_dict = deepcopy(self._app_properties)

        for command_name, command in self.commands.items():
            app_dict[command_name] = command.command

        if self.command_chain:
            app_dict["command-chain"] = self.command_chain

        if self.install_mode:
            app_dict["install-mode"] = self.install_mode

        # Adjust socket values to formats snap.yaml accepts
        sockets = app_dict.get("sockets", dict())
        for socket in sockets.values():
            mode = socket.get("socket-mode")
            if mode is not None:
                socket["socket-mode"] = yaml_utils.OctInt(mode)

        # Strip keys.
        for key in ["adapter", "desktop"]:
            if key in app_dict:
                app_dict.pop(key)

        # Apply passthrough keys.
        app_dict.update(self.passthrough)

        # Ensure passthrough is removed.
        app_dict.pop("passthrough", None)

        return app_dict

    def __repr__(self) -> str:
        return repr(self.__dict__)

    def __str__(self) -> str:
        return str(self.__dict__)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess

import testscenarios

from tests import integration


class ChecksumAlgorithmsTestCase(testscenarios.WithScenarios, integration.TestCase):

    scenarios = [
        (project_dir, {"project_dir": project_dir})
        for project_dir in [
            "checksum-algorithms",
            "deb-with-checksum",
            "rpm-with-checksum",
        ]
    ]

    def test_checksum_algorithms(self):
        self.run_snapcraft("pull", self.project_dir)


class InvalidChecksumsTestCase(testscenarios.WithScenarios, integration.TestCase):

    scenarios = [
        (part, {"part": part})
        for part in [
            "checksum-md5",
            "checksum-sha1",
            "checksum-sha224",
            "checksum-sha256",
            "checksum-sha384",
            "checksum-sha512",
            "checksum-sha3-284",
            "checksum-sha3-256",
            "checksum-sha3-512",
        ]
    ]

    def test_checksum_invalid(self):
        project_dir = "checksum-algorithms-invalid"
        self.assertRaises(
            subprocess.CalledProcessError,
            self.run_snapcraft,
            ["pull", self.part],
            project_dir,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import click

from snapcraft.internal import deprecations

from . import echo

_CMD_DEPRECATED_REPLACEMENTS = {
    "strip": "prime",
    "push": "upload",
    "push-metadata": "upload-metadata",
    "history": "list-revisions",
    "list-registered": "list",
    "registered": "list",
}

_CMD_ALIASES = {
    "keys": "list-keys",
    "revisions": "list-revisions",
    "plugins": "list-plugins",
    "collaborators": "edit-collaborators",
    "extensions": "list-extensions",
    "tracks": "list-tracks",
}

_CMD_DEPRECATION_NOTICES = {
    "history": "dn4",
    "push": "dn11",
    "push-metadata": "dn11",
    "list-registered": "dn12",
    "registered": "dn12",
}

_CMD_LEGACY = ["cleanbuild", "refresh", "search", "update", "define"]


class SnapcraftGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        new_cmd_name = _CMD_DEPRECATED_REPLACEMENTS.get(cmd_name)
        if new_cmd_name:
            if _CMD_DEPRECATION_NOTICES.get(cmd_name):
                deprecations.handle_deprecation_notice(
                    _CMD_DEPRECATION_NOTICES.get(cmd_name)
                )
            else:
                echo.warning(
                    "DEPRECATED: Use {!r} instead of {!r}".format(
                        new_cmd_name, cmd_name
                    )
                )
            cmd = click.Group.get_command(self, ctx, new_cmd_name)
        else:
            cmd_name = _CMD_ALIASES.get(cmd_name, cmd_name)
            cmd = click.Group.get_command(self, ctx, cmd_name)
        return cmd

    def list_commands(self, ctx):
        commands = super().list_commands(ctx)
        # Let's keep edit-collaborators hidden until we get the green light
        # from the store.
        commands.pop(commands.index("edit-collaborators"))

        # Hide the legacy commands
        for command in _CMD_LEGACY:
            commands.pop(commands.index(command))

        # Hide commands with unstable cli
        commands.pop(commands.index("promote"))
        commands.pop(commands.index("remote-build"))

        return commands



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This rust plugin is useful for building rust based parts.

Rust uses cargo to drive the build.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - rust-features
      (list of strings)
      Features used to build optional dependencies

    - rust-path
      (list of strings, default [.])
      Build specific workspace crates
      Only one item is currently supported.
"""

from textwrap import dedent
from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import PluginV2


class RustPlugin(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "rust-path": {
                    "type": "array",
                    "minItems": 1,
                    # TODO support more than one item.
                    "maxItems": 1,
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": ["."],
                },
                "rust-features": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": [],
                },
            },
            "required": ["source"],
        }

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        return {"curl", "gcc", "git"}

    def get_build_environment(self) -> Dict[str, str]:
        return {"PATH": "${HOME}/.cargo/bin:${PATH}"}

    def _get_rustup_command(self) -> str:
        return dedent(
            """\
        if ! command -v rustup 2>/dev/null; then
            curl --proto '=https' --tlsv1.2 -sSf https://sh.rustup.rs | sh -s -- -y --no-modify-path --profile=minimal
            export PATH="${HOME}/.cargo/bin:${PATH}"
        fi
        """
        )

    def _get_install_command(self) -> str:
        cmd = [
            "cargo",
            "install",
            "--locked",
            "--path",
            self.options.rust_path[0],
            "--root",
            '"${SNAPCRAFT_PART_INSTALL}"',
            "--force",
        ]

        if self.options.rust_features:
            cmd.extend(
                ["--features", "'{}'".format(" ".join(self.options.rust_features))]
            )

        return " ".join(cmd)

    def get_build_commands(self) -> List[str]:
        return [self._get_rustup_command(), self._get_install_command()]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from ._plugin import PluginV2  # noqa: F401

# The plugin code requires imports that are platform specific.
if sys.platform == "linux":
    from .autotools import AutotoolsPlugin  # noqa: F401
    from .catkin import CatkinPlugin  # noqa: F401
    from .catkin_tools import CatkinToolsPlugin  # noqa: F401
    from .cmake import CMakePlugin  # noqa: F401
    from .colcon import ColconPlugin  # noqa: F401
    from .conda import CondaPlugin  # noqa: F401
    from .dump import DumpPlugin  # noqa: F401
    from .go import GoPlugin  # noqa: F401
    from .make import MakePlugin  # noqa: F401
    from .meson import MesonPlugin  # noqa: F401
    from .nil import NilPlugin  # noqa: F401
    from .npm import NpmPlugin  # noqa: F401
    from .python import PythonPlugin  # noqa: F401
    from .qmake import QMakePlugin  # noqa: F401
    from .rust import RustPlugin  # noqa: F401



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testtools.matchers import Equals

from snapcraft.internal.meta import errors
from snapcraft.internal.meta.system_user import SystemUser, SystemUserScope
from tests import unit


class SystemUserTests(unit.TestCase):
    def test_valid_user(self):
        user = SystemUser(name="name", scope=SystemUserScope.SHARED)
        user.validate()

        user_dict = {"scope": "shared"}
        user = SystemUser.from_dict(user_name="name", user_dict=user_dict)

    def test_invalid_user_empty_scope(self):
        user = SystemUser(name="name", scope=None)

        error = self.assertRaises(errors.SystemUsernamesValidationError, user.validate)

        self.assertThat(error._name, Equals("name"))
        self.assertThat(error._message, Equals("scope None is invalid"))

    def test_invalid_user_empty_scope_from_dict(self):
        user_dict = {"scope": None}

        error = self.assertRaises(
            errors.SystemUsernamesValidationError,
            SystemUser.from_dict,
            user_name="name",
            user_dict=user_dict,
        )

        self.assertThat(error._name, Equals("name"))
        self.assertThat(error._message, Equals("'scope' is undefined"))

    def test_invalid_user_non_shared_scope(self):
        user = SystemUser(name="name", scope="none")

        error = self.assertRaises(errors.SystemUsernamesValidationError, user.validate)

        self.assertThat(error._name, Equals("name"))
        self.assertThat(error._message, Equals("scope 'none' is invalid"))

    def test_invalid_user_non_shared_scope_from_dict(self):
        user_dict = {"scope": "ghost"}

        error = self.assertRaises(
            errors.SystemUsernamesValidationError,
            SystemUser.from_dict,
            user_name="name",
            user_dict=user_dict,
        )

        self.assertThat(error._name, Equals("name"))
        self.assertThat(error._message, Equals("scope 'ghost' is invalid"))

    def test_from_object_dict(self):
        user_name = "name"
        user_object = {"scope": "shared"}

        user = SystemUser.from_object(user_name=user_name, user_object=user_object)

        self.assertThat(user.name, Equals(user_name))
        self.assertThat(user.scope, Equals(SystemUserScope.SHARED))

    def test_from_object_dict_invalid(self):
        user_name = "name"
        user_object = {"scope": "invalid_scope"}

        error = self.assertRaises(
            errors.SystemUsernamesValidationError,
            SystemUser.from_object,
            user_name=user_name,
            user_object=user_object,
        )

        self.assertThat(error._name, Equals("name"))
        self.assertThat(error._message, Equals("scope 'invalid_scope' is invalid"))

    def test_from_object_none(self):
        user_name = "name"
        user_object = None

        error = self.assertRaises(
            errors.SystemUsernamesValidationError,
            SystemUser.from_object,
            user_name=user_name,
            user_object=user_object,
        )

        self.assertThat(error._name, Equals("name"))
        self.assertThat(error._message, Equals("undefined user"))

    def test_from_object_string(self):
        user_name = "name"
        user_object = "shared"

        user = SystemUser.from_object(user_name=user_name, user_object=user_object)

        self.assertThat(user.name, Equals(user_name))
        self.assertThat(user.scope, Equals(SystemUserScope.SHARED))

    def test_from_object_string_invalid(self):
        user_name = "name"
        user_object = "invalid_scope"

        error = self.assertRaises(
            errors.SystemUsernamesValidationError,
            SystemUser.from_object,
            user_name=user_name,
            user_object=user_object,
        )

        self.assertThat(error._name, Equals("name"))
        self.assertThat(error._message, Equals("scope 'invalid_scope' is invalid"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from testtools import TestCase
from testtools.matchers import Equals

from snapcraft.plugins.v2.rust import RustPlugin


class RustPluginTest(TestCase):
    def test_schema(self):
        schema = RustPlugin.get_schema()

        self.assertThat(
            schema,
            Equals(
                {
                    "$schema": "http://json-schema.org/draft-04/schema#",
                    "additionalProperties": False,
                    "properties": {
                        "rust-features": {
                            "default": [],
                            "items": {"type": "string"},
                            "type": "array",
                            "uniqueItems": True,
                        },
                        "rust-path": {
                            "default": ["."],
                            "items": {"type": "string"},
                            "maxItems": 1,
                            "minItems": 1,
                            "type": "array",
                            "uniqueItems": True,
                        },
                    },
                    "required": ["source"],
                    "type": "object",
                }
            ),
        )

    def test_get_build_packages(self):
        plugin = RustPlugin(part_name="my-part", options=lambda: None)

        self.assertThat(plugin.get_build_packages(), Equals({"curl", "gcc", "git"}))

    def test_get_build_environment(self):
        plugin = RustPlugin(part_name="my-part", options=lambda: None)

        self.assertThat(
            plugin.get_build_environment(),
            Equals({"PATH": "${HOME}/.cargo/bin:${PATH}"}),
        )

    def test_get_build_commands(self):
        class Options:
            rust_channel = ""
            rust_path = ["."]
            rust_features = []

        plugin = RustPlugin(part_name="my-part", options=Options())

        self.assertThat(
            plugin.get_build_commands(),
            Equals(
                [
                    dedent(
                        """\
                    if ! command -v rustup 2>/dev/null; then
                        curl --proto '=https' --tlsv1.2 -sSf https://sh.rustup.rs | sh -s -- -y --no-modify-path --profile=minimal
                        export PATH="${HOME}/.cargo/bin:${PATH}"
                    fi
                        """
                    ),
                    'cargo install --locked --path . --root "${SNAPCRAFT_PART_INSTALL}" --force',
                ]
            ),
        )

    def test_get_install_command_with_features(self):
        class Options:
            rust_channel = ""
            rust_path = ["path"]
            rust_features = ["my-feature", "your-feature"]

        plugin = RustPlugin(part_name="my-part", options=Options())

        self.assertThat(
            plugin._get_install_command(),
            Equals(
                "cargo install --locked --path path --root \"${SNAPCRAFT_PART_INSTALL}\" --force --features 'my-feature your-feature'"
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
import stat
from typing import Union


def _executable_is_valid(path: Union[str, pathlib.Path]) -> bool:
    if not os.path.isfile(path):
        return False

    mode = os.stat(path).st_mode
    return bool(mode & stat.S_IXUSR or mode & stat.S_IXGRP or mode & stat.S_IXOTH)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Set, Union

from snapcraft import formatting_utils

# Ideally we'd just use Collection from typing, but that wasn't introduced
# until 3.6
StrCollection = Union[List[str], Set[str]]
DependencyCollection = Union[List["Dependency"], Set["Dependency"]]


class Dependency:
    def __init__(self, *, part_name, step):
        self.part_name = part_name
        self.step = step


class DirtyReport:
    """The DirtyReport class explains why a given step is dirty.

    A dirty step is defined to be a step that has run, but since doing so one
    of the following things have happened:

    - One or more YAML properties used by the step (e.g. `stage-packages`) have
      changed.

    - One of more project options (e.g. the `--target-arch` CLI option) have
      changed.

    - One of more of its dependencies have been re-staged.
    """

    def __init__(
        self,
        *,
        dirty_properties: StrCollection = None,
        dirty_project_options: StrCollection = None,
        changed_dependencies: DependencyCollection = None
    ) -> None:
        """Create a new DirtyReport.

        :param list dirty_properties: YAML properties that have changed.
        :param list dirty_project_options: Project options that have changed.
        :param list changed_dependencies: Dependencies that have changed.
        """
        self.dirty_properties = dirty_properties
        self.dirty_project_options = dirty_project_options
        self.changed_dependencies = changed_dependencies

    def get_report(self) -> str:
        """Get verbose report.

        :return: Report why the part is dirty.
        :rtype: str
        """
        messages = []  # type: List[str]

        if self.dirty_properties:
            humanized_properties = formatting_utils.humanize_list(
                self.dirty_properties, "and"
            )
            pluralized_connection = formatting_utils.pluralize(
                self.dirty_properties, "property appears", "properties appear"
            )
            messages.append(
                "The {} part {} to have changed.\n".format(
                    humanized_properties, pluralized_connection
                )
            )

        if self.dirty_project_options:
            humanized_options = formatting_utils.humanize_list(
                self.dirty_project_options, "and"
            )
            pluralized_connection = formatting_utils.pluralize(
                self.dirty_project_options, "option appears", "options appear"
            )
            messages.append(
                "The {} project {} to have changed.\n".format(
                    humanized_options, pluralized_connection
                )
            )

        if self.changed_dependencies:
            dependencies = [d.part_name for d in self.changed_dependencies]
            messages.append(
                "{} changed: {}\n".format(
                    formatting_utils.pluralize(
                        dependencies, "A dependency has", "Some dependencies have"
                    ),
                    formatting_utils.humanize_list(dependencies, "and"),
                )
            )

        return "".join(messages)

    def get_summary(self) -> str:
        """Get summarized report.

        :return: Short summary of why the part is dirty.
        :rtype: str
        """
        reasons = []

        reasons_count = 0
        if self.dirty_properties:
            reasons_count += 1
        if self.dirty_project_options:
            reasons_count += 1
        if self.changed_dependencies:
            reasons_count += 1

        if self.dirty_properties:
            # Be specific only if this is the only reason
            if reasons_count > 1 or len(self.dirty_properties) > 1:
                reasons.append("properties")
            else:
                reasons.append(
                    "{!r} property".format(next(iter(self.dirty_properties)))
                )

        if self.dirty_project_options:
            # Be specific only if this is the only reason
            if reasons_count > 1 or len(self.dirty_project_options) > 1:
                reasons.append("options")
            else:
                reasons.append(
                    "{!r} option".format(next(iter(self.dirty_project_options)))
                )

        if self.changed_dependencies:
            # Be specific only if this is the only reason
            if reasons_count > 1 or len(self.changed_dependencies) > 1:
                reasons.append("dependencies")
            else:
                reasons.append(
                    "{!r}".format(next(iter(self.changed_dependencies)).part_name)
                )

        return "{} changed".format(formatting_utils.humanize_list(reasons, "and", "{}"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import functools
import logging
import os
import pathlib
import sys

import click

import snapcraft
from snapcraft.internal import log

from ._command_group import SnapcraftGroup
from ._errors import exception_handler
from ._options import add_provider_options
from .assertions import assertionscli
from .containers import containerscli
from .discovery import discoverycli
from .extensions import extensioncli
from .help import helpcli
from .legacy import legacycli
from .lifecycle import lifecyclecli
from .remote import remotecli
from .store import storecli
from .version import SNAPCRAFT_VERSION_TEMPLATE, versioncli

logger = logging.getLogger(__name__)

command_groups = [
    storecli,
    assertionscli,
    containerscli,
    discoverycli,
    helpcli,
    legacycli,
    lifecyclecli,
    extensioncli,
    versioncli,
    remotecli,
]


def configure_requests_ca() -> None:
    """Force `requests` library to use host certificates, if found.

    Otherwise it will use the ones provided by `certifi`, which may be
    out of date or missing user-installed certificates.

    The bundle paths below should cover the majority of the Linux
    ecosystem, but OS X and Windows are currently unhandled which results
    in default behavior (use the certifi certificates).
    """
    ca_bundle_paths = [
        p
        for p in [
            pathlib.Path("/etc/ssl/certs/ca-certificates.crt"),
            pathlib.Path("/etc/pki/tls/certs/ca-bundle.crt"),
        ]
        if p.exists()
    ]

    if "REQUESTS_CA_BUNDLE" not in os.environ and ca_bundle_paths:
        os.environ["REQUESTS_CA_BUNDLE"] = str(ca_bundle_paths[0])

    if "REQUESTS_CA_BUNDLE" in os.environ:
        logger.debug(
            f"configure_requests_ca: using {os.environ['REQUESTS_CA_BUNDLE']!r} for certificate verification"
        )


@click.group(
    cls=SnapcraftGroup,
    invoke_without_command=True,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.version_option(
    message=SNAPCRAFT_VERSION_TEMPLATE, version=snapcraft.__version__  # type: ignore
)
@click.pass_context
@add_provider_options(hidden=True)
@click.option("--debug", "-d", is_flag=True)
def run(ctx, debug, catch_exceptions=False, **kwargs):
    """Snapcraft is a delightful packaging tool."""

    is_snapcraft_developer_debug = kwargs["enable_developer_debug"]
    if is_snapcraft_developer_debug:
        log_level = logging.DEBUG
        click.echo(
            "Starting snapcraft {} from {}.".format(
                snapcraft.__version__, os.path.dirname(__file__)
            )
        )
    else:
        log_level = logging.INFO

    # Setup global exception handler (to be called for unhandled exceptions)
    sys.excepthook = functools.partial(
        exception_handler, debug=is_snapcraft_developer_debug
    )

    # In an ideal world, this logger setup would be replaced
    log.configure(log_level=log_level)

    # Configure `requests` library to use system certificates.
    configure_requests_ca()

    # Payload information about argv
    ctx.obj = dict(argv=sys.argv)

    # The default command
    if not ctx.invoked_subcommand:
        snap_command = lifecyclecli.commands["snap"]
        # Fill in the context with default values for the snap command.
        if "directory" not in ctx.params:
            ctx.params["directory"] = None
        if "output" not in ctx.params:
            ctx.params["output"] = None
        snap_command.invoke(ctx)


# This would be much easier if they were subcommands
for command_group in command_groups:
    for command in command_group.commands:
        run.add_command(command_group.commands[command])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
from unittest import mock

import fixtures
import pytest
from testtools.matchers import Equals, HasLength

from snapcraft.internal import errors
from snapcraft.internal.meta.snap import Snap
from snapcraft.plugins.v1 import ant
from snapcraft.project import Project
from tests import unit

from . import PluginsV1BaseTestCase


class AntPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        schema = ant.AntPlugin.schema()

        properties = schema["properties"]
        for expected in ["ant-properties", "ant-build-targets"]:
            self.assertTrue(
                expected in properties,
                "Expected {!r} to be included in properties".format(expected),
            )

        properties_type = schema["properties"]["ant-properties"]["type"]
        self.assertThat(
            properties_type,
            Equals("object"),
            'Expected "ant-properties" "type" to be "object", '
            'but it was "{}"'.format(properties_type),
        )
        build_targets_type = schema["properties"]["ant-build-targets"]["type"]
        self.assertThat(
            build_targets_type,
            Equals("array"),
            'Expected "ant-build-targets" "type" to be "object", '
            'but it was "{}"'.format(build_targets_type),
        )

    def test_get_pull_properties(self):
        expected_pull_properties = [
            "ant-channel",
            "ant-version",
            "ant-version-checksum",
            "ant-openjdk-version",
        ]
        resulting_pull_properties = ant.AntPlugin.get_pull_properties()

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, resulting_pull_properties)

    def test_get_build_properties(self):
        expected_build_properties = [
            "ant-build-targets",
            "ant-properties",
            "ant-buildfile",
        ]
        resulting_build_properties = ant.AntPlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)


def _get_expected_java_version(ant_plugin) -> str:
    base = ant_plugin.project._snap_meta.base
    ant_openjdk_version = ant_plugin.options.ant_openjdk_version

    if ant_openjdk_version:
        expected_java_version = ant_openjdk_version
    elif base in ("core", "core16"):
        expected_java_version = "9"
    else:
        expected_java_version = "11"

    return expected_java_version


_BASE_JAVA_COMBINATIONS = [
    ("", "core"),
    ("8", "core"),
    ("", "core16"),
    ("8", "core16"),
    ("", "core18"),
    ("11", "core18"),
]


@pytest.fixture(params=_BASE_JAVA_COMBINATIONS)
def ant_plugin(tmp_work_path, request):
    """Return an instance of AntPlugin setup with different bases and java versions."""
    java_version, base = request.param

    class Options:
        ant_properties = {}
        ant_build_targets = None
        ant_channel = None
        ant_version = "1.10.5"
        ant_version_checksum = "sha512/a7f1e0cec9d5ed1b3ab6cddbb9364f127305a997bbc88ecd734f9ef142ec0332375e01ace3592759bb5c3307cd9c1ac0a78a30053f304c7030ea459498e4ce4e"
        ant_openjdk_version = java_version
        ant_buildfile = None

    project = Project()
    project._snap_meta = Snap(name="test-snap", base=base, confinement="strict")

    return ant.AntPlugin("test-part", Options(), project)


@pytest.fixture()
def ant_plugin_with_assets(ant_plugin):
    """Return an instance of AntPlugin with artifacts setup using ant_plugin."""
    pathlib.Path(ant_plugin.sourcedir).mkdir(parents=True)

    expected_java_version = _get_expected_java_version(ant_plugin)

    fake_java_path = (
        pathlib.Path(ant_plugin.installdir)
        / f"usr/lib/jvm/java-{expected_java_version}-openjdk-amd64/bin/java"
    )
    fake_java_path.parent.mkdir(parents=True)
    fake_java_path.touch()

    ant_tar_path = (
        pathlib.Path(ant_plugin.partdir)
        / f"ant/apache-ant-{ant_plugin.options.ant_version}-bin.tar.bz2"
    )
    ant_tar_path.parent.mkdir(parents=True)
    ant_tar_path.touch()

    return ant_plugin


def test_stage_and_build_packages(ant_plugin):
    expected_java_version = _get_expected_java_version(ant_plugin)

    assert ant_plugin.stage_packages == [
        f"openjdk-{expected_java_version}-jre-headless"
    ]
    assert ant_plugin.build_packages == [
        f"openjdk-{expected_java_version}-jdk-headless"
    ]


def test_build(mock_tar, mock_run, ant_plugin_with_assets):
    plugin = ant_plugin_with_assets

    def fake_run(cmd, *args, **kwargs):
        os.makedirs(os.path.join(plugin.builddir, "target"))
        open(os.path.join(plugin.builddir, "target", "dummy.jar"), "w").close()

    mock_run.side_effect = fake_run

    plugin.build()

    mock_run.assert_called_once_with(["ant"], cwd=plugin.builddir, env=mock.ANY)
    mock_tar.assert_called_once_with(
        ant._ANT_ARCHIVE_FORMAT_URL.format(version=plugin.options.ant_version),
        mock.ANY,
        source_checksum=plugin.options.ant_version_checksum,
    )


def test_build_with_options(mock_tar, mock_run, ant_plugin_with_assets):
    plugin = ant_plugin_with_assets
    plugin.options.ant_build_targets = ["artifacts", "jar"]
    plugin.options.ant_properties = {"basedir": "."}

    plugin.build()

    mock_run.assert_called_once_with(
        ["ant", "artifacts", "jar", "-Dbasedir=."], cwd=plugin.builddir, env=mock.ANY
    )
    mock_tar.assert_called_once_with(
        ant._ANT_ARCHIVE_FORMAT_URL.format(version=plugin.options.ant_version),
        mock.ANY,
        source_checksum=plugin.options.ant_version_checksum,
    )


def test_build_with_explicit_buildfile(mock_tar, mock_run, ant_plugin_with_assets):
    plugin = ant_plugin_with_assets
    plugin.options.ant_buildfile = "test.xml"

    plugin.build()

    mock_run.assert_called_once_with(
        ["ant", "-f", "test.xml"], cwd=plugin.builddir, env=mock.ANY
    )
    mock_tar.assert_called_once_with(
        ant._ANT_ARCHIVE_FORMAT_URL.format(version=plugin.options.ant_version),
        mock.ANY,
        source_checksum=plugin.options.ant_version_checksum,
    )


class AntPluginSnapTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class Options:
            ant_properties = {}
            ant_build_targets = None
            ant_channel = None
            ant_version = None
            ant_version_checksum = None
            ant_openjdk_version = None
            ant_buildfile = None

        self.options = Options()

        self.run_mock = self.useFixture(
            fixtures.MockPatch("snapcraft.internal.common.run")
        ).mock
        self.tar_mock = self.useFixture(
            fixtures.MockPatch("snapcraft.internal.sources.Tar")
        ).mock

    def create_assets(self, plugin):
        os.makedirs(plugin.sourcedir)

        fake_java_path = os.path.join(
            plugin.installdir,
            "usr",
            "lib",
            "jvm",
            "java-{}-openjdk-amd64".format(plugin._java_version),
            "bin",
            "java",
        )
        os.makedirs(os.path.dirname(fake_java_path))
        open(fake_java_path, "w").close()

    def test_build_with_default_snap(self):
        plugin = ant.AntPlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        def side(l, **kwargs):
            os.makedirs(os.path.join(plugin.builddir, "target"))
            open(os.path.join(plugin.builddir, "target", "dummy.jar"), "w").close()

        self.run_mock.side_effect = side

        plugin.build()

        self.tar_mock.assert_not_called()
        self.run_mock.assert_called_once_with(
            ["ant"], cwd=plugin.builddir, env=mock.ANY
        )
        self.assertEqual(plugin.build_snaps, ["ant/" + ant._DEFAULT_ANT_SNAP_CHANNEL])

    def test_build_with_explicit_snap(self):
        self.options.ant_channel = "other/channel"
        plugin = ant.AntPlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        def side(l, **kwargs):
            os.makedirs(os.path.join(plugin.builddir, "target"))
            open(os.path.join(plugin.builddir, "target", "dummy.jar"), "w").close()

        self.run_mock.side_effect = side

        plugin.build()

        self.tar_mock.assert_not_called()
        self.run_mock.assert_called_once_with(
            ["ant"], cwd=plugin.builddir, env=mock.ANY
        )
        self.assertEqual(plugin.build_snaps, ["ant/other/channel"])


class AntPluginUnsupportedBase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "unsupported-base"

        class Options:
            ant_properties = {}
            ant_build_targets = None
            ant_channel = None
            ant_version = "1.10.5"
            ant_version_checksum = "sha512/a7f1e0cec9d5ed1b3ab6cddbb9364f127305a997bbc88ecd734f9ef142ec0332375e01ace3592759bb5c3307cd9c1ac0a78a30053f304c7030ea459498e4ce4e"
            ant_openjdk_version = ""

        self.options = Options()

    def test_unsupported_base_raises(self):
        self.assertRaises(
            errors.PluginBaseError,
            ant.AntPlugin,
            "test-part",
            self.options,
            self.project,
        )


class TestUnsupportedJDKVersionError:

    scenarios = (
        (
            "core",
            dict(
                base="core",
                version="11",
                expected_message=(
                    "The ant-openjdk-version plugin property was set to '11'.\n"
                    "Valid values for the 'core' base are: '8' or '9'."
                ),
            ),
        ),
        (
            "core16",
            dict(
                base="core16",
                version="11",
                expected_message=(
                    "The ant-openjdk-version plugin property was set to '11'.\n"
                    "Valid values for the 'core16' base are: '8' or '9'."
                ),
            ),
        ),
        (
            "core18",
            dict(
                base="core18",
                version="9",
                expected_message=(
                    "The ant-openjdk-version plugin property was set to '9'.\n"
                    "Valid values for the 'core18' base are: '11' or '8'."
                ),
            ),
        ),
    )

    def test_use_invalid_openjdk_version_fails(self, base, version, expected_message):
        class Options:
            ant_properties = {}
            ant_build_targets = None
            ant_channel = None
            ant_version = "1.10.5"
            ant_version_checksum = "sha512/a7f1e0cec9d5ed1b3ab6cddbb9364f127305a997bbc88ecd734f9ef142ec0332375e01ace3592759bb5c3307cd9c1ac0a78a30053f304c7030ea459498e4ce4e"
            ant_openjdk_version = version

        project = Project()
        project._snap_meta = Snap(name="test-snap", base=base, confinement="strict")

        with pytest.raises(ant.UnsupportedJDKVersionError) as error:
            ant.AntPlugin("test-part", Options(), project)
            assert str(error) == expected_message



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015, 2018-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This plugin is useful for building parts that use maven.

The maven build system is commonly used to build Java projects.
The plugin requires a pom.xml in the root of the source tree.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - maven-options:
      (list of strings)
      Flags to pass to the build using the maven semantics for parameters.

    - maven-targets:
      (list of strings)
      List of target directories where the built jars are stored.
      It corresponds to the <build><directory>target</directory></build>
      from the Maven's build declaration.

    - maven-version:
      (string)
      The version of maven you want to use to build the source artifacts.
      Defaults to the current release downloadable from
      https://archive.apache.org/dist/maven/maven-3/.

    - maven-version-checksum:
      (string)
      The checksum for maven-version in the form of <digest-type>/<digest>.
      As an example "sha512/2a803f578f341e164f6753e410413d16ab60fab...".

    - maven-openjdk-version:
      (string)
      openjdk version available to the base to use. If not set the latest
      version available to the base will be used.
"""

import logging
import os
import shutil
import textwrap
from glob import glob
from typing import Optional, Sequence
from urllib.parse import urlparse
from xml.etree import ElementTree

from snapcraft import file_utils, formatting_utils
from snapcraft.internal import errors, sources
from snapcraft.plugins.v1 import PluginV1

logger = logging.getLogger(__name__)

_DEFAULT_MAVEN_VERSION = "3.5.4"
_DEFAULT_MAVEN_CHECKSUM = "sha512/2a803f578f341e164f6753e410413d16ab60fabe31dc491d1fe35c984a5cce696bc71f57757d4538fe7738be04065a216f3ebad4ef7e0ce1bb4c51bc36d6be86"
_MAVEN_URL = "https://archive.apache.org/dist/maven/maven-3/{version}/binaries/apache-maven-{version}-bin.tar.gz"


class EmptyBuildTargetDirectoryError(errors.SnapcraftException):
    def __init__(self, target_directory: str) -> None:
        self._target_directory = target_directory

    def get_details(self) -> Optional[str]:
        return textwrap.dedent(
            """\
            This could happen if your `maven-targets` points to a directory that is not the same as the one
            in your Maven's <build><directory>target</directory></build> declaration. By default, Snap's
            Maven plugin won't need `maven-targets` declaration as it will work with the default Maven behavior.
            If you are not sure, try removing `maven-targets` definition and build your Snap again."""
        )

    def get_brief(self) -> str:
        return "Could not find built jar files for part in the following directory: {!r}".format(
            self._target_directory
        )

    def get_resolution(self) -> str:
        return textwrap.dedent(
            """\
            `maven-targets` should point the same directory as Maven's
            <build><directory>target</directory></build> declaration."""
        )


class UnsupportedJDKVersionError(errors.SnapcraftError):
    fmt = (
        "The maven-openjdk-version plugin property was set to {version!r}.\n"
        "Valid values for the {base!r} base are: {valid_versions}."
    )

    def __init__(
        self, *, base: str, version: str, valid_versions: Sequence[str]
    ) -> None:
        super().__init__(
            base=base,
            version=version,
            valid_versions=formatting_utils.humanize_list(
                valid_versions, conjunction="or"
            ),
        )


class MavenPlugin(PluginV1):
    @classmethod
    def schema(cls):

        schema = super().schema()

        schema["properties"]["maven-options"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }

        schema["properties"]["maven-targets"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [""],
        }

        schema["properties"]["maven-version"] = {"type": "string"}

        schema["properties"]["maven-version-checksum"] = {"type": "string"}

        schema["properties"]["maven-openjdk-version"] = {
            "type": "string",
            "default": "",
        }

        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the pull step dirty.
        return ["maven-version", "maven-version-checksum", "maven-openjdk-version"]

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["maven-options", "maven-targets"]

    @property
    def _maven_tar(self):
        if self._maven_tar_handle is None:
            maven_uri = _MAVEN_URL.format(version=self._maven_version)
            self._maven_tar_handle = sources.Tar(
                maven_uri, self._maven_dir, source_checksum=self._maven_checksum
            )
        return self._maven_tar_handle

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        self._setup_maven()
        self._setup_base_tools(project._get_build_base())

    def _setup_base_tools(self, base):
        if base in ("core", "core16"):
            valid_versions = ["8", "9"]
        elif base == "core18":
            valid_versions = ["8", "11"]

        version = self.options.maven_openjdk_version
        if not version:
            version = valid_versions[-1]
        elif version not in valid_versions:
            raise UnsupportedJDKVersionError(
                version=version, base=base, valid_versions=valid_versions
            )

        self.stage_packages.append("openjdk-{}-jre-headless".format(version))
        self.build_packages.append("openjdk-{}-jdk-headless".format(version))
        self._java_version = version

    def _setup_maven(self):
        self._maven_tar_handle = None
        self._maven_dir = os.path.join(self.partdir, "maven")
        if self.options.maven_version:
            self._maven_version = self.options.maven_version
            self._maven_checksum = self.options.maven_version_checksum
        else:
            self._maven_version = _DEFAULT_MAVEN_VERSION
            self._maven_checksum = _DEFAULT_MAVEN_CHECKSUM

    def _use_proxy(self):
        return any(k in os.environ for k in ("http_proxy", "https_proxy"))

    def pull(self):
        super().pull()

        os.makedirs(self._maven_dir, exist_ok=True)
        self._maven_tar.download()

    def clean_pull(self):
        super().clean_pull()
        # Remove the maven directory (if any)
        if os.path.exists(self._maven_dir):
            shutil.rmtree(self._maven_dir)

    def build(self):
        super().build()

        self._maven_tar.provision(
            self._maven_dir, clean_target=False, keep_tarball=True
        )

        mvn_cmd = ["mvn", "package"]
        if self._use_proxy():
            settings_path = os.path.join(self.partdir, "m2", "settings.xml")
            _create_settings(settings_path)
            mvn_cmd += ["-s", settings_path]

        self.run(mvn_cmd + self.options.maven_options, rootdir=self.builddir)

        for f in self.options.maven_targets:
            src = os.path.join(self.builddir, f, "target")
            jarfiles = glob(os.path.join(src, "*.jar"))
            warfiles = glob(os.path.join(src, "*.war"))
            arfiles = glob(os.path.join(src, "*.[jw]ar"))

            if len(arfiles) == 0:
                raise EmptyBuildTargetDirectoryError(src)
            if len(jarfiles) > 0 and len(f) == 0:
                basedir = "jar"
            elif len(warfiles) > 0 and len(f) == 0:
                basedir = "war"
            else:
                basedir = f

            targetdir = os.path.join(self.installdir, basedir)
            os.makedirs(targetdir, exist_ok=True)
            for src in arfiles:
                # Make rebuilding easier when enabled
                base = os.path.basename(src)
                dst = os.path.join(targetdir, base)
                if os.path.exists(dst):
                    os.unlink(dst)
                file_utils.link_or_copy(src, dst)

        self._create_symlinks()

    def _create_symlinks(self):
        if self.project._get_build_base() not in ("core18", "core16", "core"):
            raise errors.PluginBaseError(
                part_name=self.name, base=self.project._get_build_base()
            )

        os.makedirs(os.path.join(self.installdir, "bin"), exist_ok=True)
        java_bin = glob(
            os.path.join(
                self.installdir,
                "usr",
                "lib",
                "jvm",
                "java-{}-openjdk-*".format(self._java_version),
                "bin",
                "java",
            )
        )[0]
        os.symlink(
            os.path.relpath(java_bin, os.path.join(self.installdir, "bin")),
            os.path.join(self.installdir, "bin", "java"),
        )

    def run(self, cmd, rootdir):
        super().run(cmd, cwd=rootdir, env=self._build_environment())

    def _build_environment(self):
        env = os.environ.copy()
        maven_bin = os.path.join(self._maven_dir, "bin")

        if env.get("PATH"):
            new_path = "{}:{}".format(maven_bin, env.get("PATH"))
        else:
            new_path = maven_bin

        env["PATH"] = new_path
        return env


def _create_settings(settings_path):
    settings = ElementTree.Element(
        "settings",
        attrib={
            "xmlns": "http://maven.apache.org/SETTINGS/1.0.0",
            "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            "xsi:schemaLocation": (
                "http://maven.apache.org/SETTINGS/1.0.0 "
                "http://maven.apache.org/xsd/settings-1.0.0.xsd"
            ),
        },
    )
    element = ElementTree.Element("interactiveMode")
    element.text = "false"
    settings.append(element)
    proxies = ElementTree.Element("proxies")
    for protocol in ("http", "https"):
        env_name = "{}_proxy".format(protocol)
        if env_name not in os.environ:
            continue
        proxy_url = urlparse(os.environ[env_name])
        proxy = ElementTree.Element("proxy")
        proxy_tags = [
            ("id", env_name),
            ("active", "true"),
            ("protocol", protocol),
            ("host", proxy_url.hostname),
            ("port", str(proxy_url.port)),
        ]
        if proxy_url.username is not None:
            proxy_tags.extend(
                [("username", proxy_url.username), ("password", proxy_url.password)]
            )
        proxy_tags.append(("nonProxyHosts", _get_no_proxy_string()))
        for tag, text in proxy_tags:
            element = ElementTree.Element(tag)
            element.text = text
            proxy.append(element)
        proxies.append(proxy)
    settings.append(proxies)
    tree = ElementTree.ElementTree(settings)
    os.makedirs(os.path.dirname(settings_path), exist_ok=True)
    with open(settings_path, "w") as f:
        tree.write(f, encoding="unicode")
        f.write("\n")


def _get_no_proxy_string():
    no_proxy = [k.strip() for k in os.environ.get("no_proxy", "localhost").split(",")]
    return "|".join(no_proxy)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from textwrap import dedent

import fixtures
from testtools.matchers import Contains, Equals

import snapcraft.storeapi.errors
from snapcraft import storeapi

from . import FakeStoreCommandsBaseTestCase


class CloseCommandTestCase(FakeStoreCommandsBaseTestCase):
    def setUp(self):
        super().setUp()

        self.useFixture(
            fixtures.MockPatchObject(
                storeapi._dashboard_api.DashboardAPI,
                "close_channels",
                return_value=(list(), dict()),
            )
        )

    def test_close_missing_permission(self):
        self.fake_store_account_info.mock.return_value = {
            "account_id": "abcd",
            "snaps": {},
        }

        raised = self.assertRaises(
            snapcraft.storeapi.errors.StoreChannelClosingPermissionError,
            self.run_command,
            ["close", "foo", "beta"],
        )

        self.assertThat(
            str(raised),
            Equals(
                "Your account lacks permission to close channels for this snap. "
                "Make sure the logged in account has upload permissions on "
                "'foo' in series '16'."
            ),
        )

    def test_close(self):
        result = self.run_command(["close", "snap-test", "2.1/candidate"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains(
                dedent(
                    """\
            Track    Arch    Channel    Version    Revision
            2.1      amd64   stable     -          -
                             candidate  -          -
                             beta       10         19
                             edge       ↑          ↑

            The 2.1/candidate channel is now closed."""
                )
            ),
        )

    def test_close_no_revisions(self):
        self.channel_map.channel_map = list()

        result = self.run_command(["close", "snap-test", "2.1/candidate"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output.strip(), Equals("The 2.1/candidate channel is now closed.")
        )

    def test_close_multiple_channels(self):
        result = self.run_command(["close", "snap-test", "2.1/stable", "2.1/edge/test"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains(
                dedent(
                    """\
            Track    Arch    Channel    Version    Revision
            2.1      amd64   stable     -          -
                             candidate  -          -
                             beta       10         19
                             edge       ↑          ↑

            The 2.1/stable and 2.1/edge/test channels are now closed."""
                )
            ),
        )
